import logging
import math
from pathlib import Path

import numpy as np
import pandas as pd
import ruptures as rpt

from wind_up.constants import (
    RAW_POWER_COL,
    RAW_YAWDIR_COL,
    REANALYSIS_WD_COL,
    TIMESTAMP_COL,
    WINDFARM_YAWDIR_COL,
)
from wind_up.math_funcs import circ_diff
from wind_up.models import PlotConfig, WindUpConfig
from wind_up.northing import (
    add_wf_yawdir,
    apply_northing_corrections,
    calc_northed_col_name,
    check_wtg_northing,
)
from wind_up.northing_utils import YAW_OK_PW_FRACTION, add_ok_yaw_col
from wind_up.plots.optimize_northing_plots import (
    plot_diff_to_north_ref_wd,
    plot_wf_yawdir_and_reanalysis_timeseries,
    plot_yaw_diff_vs_power,
)

logger = logging.getLogger(__name__)

DECAY_FRACTION = 0.4


def northing_score_changepoint_component(changepoint_count: int) -> float:
    return float(changepoint_count)


def northing_score(
    wtg_df: pd.DataFrame, *, north_ref_wd_col: str, changepoint_count: int, rated_power: float, timebase_s: int
) -> float:
    # this component penalizes long-ish, large north errors
    max_component = max(0, wtg_df[f"long_rolling_diff_to_{north_ref_wd_col}"].abs().max() - 4) ** 2

    # this component penalizes the median north error of filtered data being far from 0
    median_component = max(0, abs(wtg_df[f"filt_diff_to_{north_ref_wd_col}"].median()) - 0.1) ** 2

    # this component penalizes raw data having any north errors
    max_weight = rated_power * YAW_OK_PW_FRACTION
    min_weight = max_weight / 1000 * (min(600, timebase_s) / 600)
    raw_wmean_component = (
        wtg_df[f"yaw_diff_to_{north_ref_wd_col}"].clip(lower=-30, upper=30)
        * wtg_df[RAW_POWER_COL].clip(lower=min_weight, upper=max_weight)
    ).abs().mean() / max_weight

    # this component encourages the correction list to be as short as possible
    changepoint_component = northing_score_changepoint_component(changepoint_count)

    return max_component + median_component + raw_wmean_component + changepoint_component


def add_northing_ok_and_diff_cols(wtg_df: pd.DataFrame, *, north_ref_wd_col: str, northed_col: str) -> pd.DataFrame:
    wtg_df = wtg_df.copy()

    wtg_df[f"yaw_diff_to_{north_ref_wd_col}"] = circ_diff(wtg_df[northed_col], wtg_df[north_ref_wd_col])
    wtg_df[f"filt_diff_to_{north_ref_wd_col}"] = wtg_df[f"yaw_diff_to_{north_ref_wd_col}"]
    wtg_df.loc[~wtg_df[f"ok_for_{north_ref_wd_col}_northing"], f"filt_diff_to_{north_ref_wd_col}"] = pd.NA
    return wtg_df


def add_northed_ok_diff_and_rolling_cols(
    wtg_df: pd.DataFrame,
    *,
    north_ref_wd_col: str,
    timebase_s: int,
    north_offset: float | None = None,
    north_offset_df: pd.DataFrame | None = None,
) -> pd.DataFrame:
    wtg_df = wtg_df.copy()
    northed_col = calc_northed_col_name(north_ref_wd_col)
    if north_offset is not None:
        wtg_df[northed_col] = wtg_df[RAW_YAWDIR_COL] + north_offset
        wtg_df[northed_col] = wtg_df[northed_col] % 360
    elif north_offset_df is not None:
        for ts, no in zip(
            north_offset_df[TIMESTAMP_COL].to_list(),
            north_offset_df["north_offset"].to_list(),
            strict=True,
        ):
            wtg_df.loc[wtg_df.index >= ts, northed_col] = wtg_df.loc[wtg_df.index >= ts, RAW_YAWDIR_COL] + no
        wtg_df[northed_col] = wtg_df[northed_col] % 360

    wtg_df = add_northing_ok_and_diff_cols(wtg_df, north_ref_wd_col=north_ref_wd_col, northed_col=northed_col)
    rolling_hours = 6
    rows_per_hour = 3600 / timebase_s
    wtg_df[f"short_rolling_diff_to_{north_ref_wd_col}"] = (
        wtg_df[f"filt_diff_to_{north_ref_wd_col}"]
        .rolling(
            center=True,
            window=round(rolling_hours * rows_per_hour),
            min_periods=round(rolling_hours * rows_per_hour // 3),
        )
        .median()
    )
    rolling_hours = 15 * 24
    wtg_df[f"long_rolling_diff_to_{north_ref_wd_col}"] = (
        wtg_df[f"filt_diff_to_{north_ref_wd_col}"]
        .rolling(
            center=True,
            window=round(rolling_hours * rows_per_hour),
            min_periods=round(rolling_hours * rows_per_hour // 3),
        )
        .median()
    )
    return wtg_df


def calc_good_north_offset(section_df: pd.DataFrame, north_ref_wd_col: str) -> float:
    return -section_df[f"filt_diff_to_{north_ref_wd_col}"].median()


def calc_north_offset_col(
    wtg_north_table: pd.DataFrame,
    *,
    wtg_df: pd.DataFrame,
    north_ref_wd_col: str,
    timebase_s: int,
) -> pd.DataFrame:
    wtg_df = wtg_df.copy()
    wtg_df = add_northing_ok_and_diff_cols(wtg_df, north_ref_wd_col=north_ref_wd_col, northed_col=RAW_YAWDIR_COL)

    wtg_north_table = wtg_north_table.copy()
    wtg_north_table = wtg_north_table.sort_values(by=[TIMESTAMP_COL], ascending=True)
    end_dt = wtg_df.index.max() + pd.Timedelta(seconds=timebase_s)
    for start_dt in reversed(wtg_north_table[TIMESTAMP_COL].to_list()):
        section_df = wtg_df[(wtg_df.index >= start_dt) & (wtg_df.index < end_dt)]
        wtg_north_table.loc[
            wtg_north_table[TIMESTAMP_COL] == start_dt,
            "north_offset",
        ] = calc_good_north_offset(
            section_df,
            north_ref_wd_col=north_ref_wd_col,
        )
        end_dt = start_dt
    return wtg_north_table


def north_table_is_valid(
    wtg_north_table: pd.DataFrame,
    *,
    wtg_df: pd.DataFrame,
    check_north_offset: bool = False,
) -> bool:
    is_valid = True
    if wtg_north_table[TIMESTAMP_COL].isna().any():
        is_valid = False
    if wtg_north_table[TIMESTAMP_COL].duplicated().any():
        is_valid = False
    if not wtg_north_table[TIMESTAMP_COL].is_monotonic_increasing:
        is_valid = False
    if wtg_north_table[TIMESTAMP_COL].min() != wtg_df.index.min():
        is_valid = False
    if wtg_north_table[TIMESTAMP_COL].max() > wtg_df.index.max():
        is_valid = False
    if check_north_offset:
        if wtg_north_table["north_offset"].isna().any():
            is_valid = False
        max_abs_north_offset = 180
        if not all(wtg_north_table["north_offset"].abs() <= max_abs_north_offset):
            is_valid = False

    return is_valid


def score_wtg_north_table(
    *,
    wtg_north_table: pd.DataFrame,
    wtg_df: pd.DataFrame,
    improve_north_offset_col: bool,
    north_ref_wd_col: str,
    rated_power: float,
    timebase_s: int,
) -> tuple[pd.DataFrame, float, pd.DataFrame]:
    if improve_north_offset_col:
        output_north_table = calc_north_offset_col(
            wtg_north_table,
            wtg_df=wtg_df,
            north_ref_wd_col=north_ref_wd_col,
            timebase_s=timebase_s,
        )
    else:
        output_north_table = wtg_north_table.copy()
    output_wtg_df = add_northed_ok_diff_and_rolling_cols(
        wtg_df.copy(),
        north_ref_wd_col=north_ref_wd_col,
        north_offset_df=output_north_table,
        timebase_s=timebase_s,
    )

    score = northing_score(
        output_wtg_df,
        north_ref_wd_col=north_ref_wd_col,
        changepoint_count=len(output_north_table),
        rated_power=rated_power,
        timebase_s=timebase_s,
    )

    return output_north_table, score, output_wtg_df


def calc_max_changepoints_to_add(changepoint_count: int, *, score: float) -> int:
    max_changepoints_to_add = 0
    headroom = score - northing_score_changepoint_component(changepoint_count + 1)
    while headroom > 0:
        max_changepoints_to_add += 1
        headroom = score - northing_score_changepoint_component(max_changepoints_to_add + 1)
    return max_changepoints_to_add


def list_possible_moves(
    prev_best_north_table: pd.DataFrame,
    *,
    do_changepoint_moves: bool,
    max_changepoints_to_add: int,
) -> list[str]:
    moves = []
    moves.extend([f"shift_changepoint_{x}_forward" for x in prev_best_north_table.index[1:]])
    moves.extend([f"shift_changepoint_{x}_back" for x in prev_best_north_table.index[1:]])
    if do_changepoint_moves:
        moves.append("add_1_changepoint")
        moves.extend([f"add_{x}_changepoints" for x in range(2, max_changepoints_to_add + 1)])
    return moves


def get_changepoint_objects(
    *,
    prev_best_wtg_df: pd.DataFrame,
    north_ref_wd_col: str,
) -> tuple[rpt.base.BaseEstimator, np.ndarray]:
    col = f"filt_diff_to_{north_ref_wd_col}"
    model = "l1"
    dropna_df = prev_best_wtg_df.dropna(subset=[col])
    signal = dropna_df[col].to_numpy()
    timestamps = dropna_df.index.to_numpy()
    algo = rpt.BottomUp(model=model).fit(signal)
    return algo, timestamps


def make_move(
    move: str,
    *,
    prev_best_north_table: pd.DataFrame,
    shift_step_size: int,
    do_changepoint_moves: bool,
    algo: rpt.base.BaseEstimator,
    timestamps: np.ndarray,
    timebase_s: int,
) -> pd.DataFrame:
    if move.startswith("shift_changepoint_"):
        cp_idx = int(move.split("_")[-2])
        sign = 1 if move.endswith("forward") else -1
        this_north_table = prev_best_north_table.copy()
        this_north_table.loc[cp_idx, TIMESTAMP_COL] = this_north_table.loc[
            cp_idx,
            TIMESTAMP_COL,
        ] + pd.Timedelta(seconds=sign * timebase_s * shift_step_size)
    elif do_changepoint_moves and move.startswith("add_"):
        n_changepoints_to_add = int(move.split("_")[1])
        bkp_idxs = algo.predict(n_bkps=n_changepoints_to_add)[:-1]
        if len(bkp_idxs) != n_changepoints_to_add:
            msg = f"found {len(bkp_idxs)} bkp_idxs, expected {n_changepoints_to_add}"
            raise RuntimeError(msg)
        dt_list = list(timestamps[bkp_idxs])

        this_north_offset_df_dtidx = prev_best_north_table.set_index(TIMESTAMP_COL)
        this_index = this_north_offset_df_dtidx.index.append(pd.Index(dt_list))
        this_north_table = this_north_offset_df_dtidx.reindex(this_index).sort_index().reset_index()
        this_north_table = this_north_table.rename(columns={"index": TIMESTAMP_COL})
    else:
        msg = f"invalid move {move}"
        raise RuntimeError(msg)
    return this_north_table


def make_move_and_score_wtg_north_table(
    move: str,
    *,
    prev_best_north_table: pd.DataFrame,
    wtg_df: pd.DataFrame,
    shift_step_size: int,
    north_ref_wd_col: str,
    rated_power: float,
    timebase_s: int,
    do_changepoint_moves: bool,
    algo: rpt.base.BaseEstimator,
    timestamps: np.ndarray,
) -> tuple[pd.DataFrame, float, pd.DataFrame]:
    this_north_table = make_move(
        move,
        prev_best_north_table=prev_best_north_table,
        shift_step_size=shift_step_size,
        do_changepoint_moves=do_changepoint_moves,
        algo=algo,
        timestamps=timestamps,
        timebase_s=timebase_s,
    )
    if north_table_is_valid(this_north_table, wtg_df=wtg_df):
        min_step_size_for_run_optimize = 100
        run_optimize_north_offset = (
            (len(this_north_table) > len(prev_best_north_table))
            or (shift_step_size > min_step_size_for_run_optimize)
            or (shift_step_size == 1)
        )
    else:
        this_north_table = prev_best_north_table
        run_optimize_north_offset = False
    this_north_table, this_score, this_wtg_df = score_wtg_north_table(
        wtg_north_table=this_north_table,
        wtg_df=wtg_df,
        improve_north_offset_col=run_optimize_north_offset,
        north_ref_wd_col=north_ref_wd_col,
        rated_power=rated_power,
        timebase_s=timebase_s,
    )
    return this_north_table, this_score, this_wtg_df


def clip_wtg_north_table(initial_wtg_north_table: pd.DataFrame, *, wtg_df: pd.DataFrame) -> pd.DataFrame:
    clipped_wtg_north_table = initial_wtg_north_table.copy()

    if clipped_wtg_north_table[TIMESTAMP_COL].min() < wtg_df.index.min():
        first_row_before_wf_df = (
            clipped_wtg_north_table.loc[clipped_wtg_north_table[TIMESTAMP_COL] <= wtg_df.index.min()]
            .sort_values(by=[TIMESTAMP_COL], ascending=False)
            .iloc[:1]
        )
        clipped_wtg_north_table = pd.concat(
            [
                first_row_before_wf_df,
                clipped_wtg_north_table[clipped_wtg_north_table[TIMESTAMP_COL] > wtg_df.index.min()],
            ],
        ).reset_index(drop=True)
        clipped_wtg_north_table.loc[0, TIMESTAMP_COL] = wtg_df.index.min()

    if clipped_wtg_north_table[TIMESTAMP_COL].min() > wtg_df.index.min():
        clipped_wtg_north_table.loc[0, TIMESTAMP_COL] = wtg_df.index.min()

    return clipped_wtg_north_table


def prep_for_optimize_wtg_north_table(
    wtg_df: pd.DataFrame,
    *,
    wtg_name: str,
    north_ref_wd_col: str,
    rated_power: float,
    timebase_s: int,
    plot_cfg: PlotConfig | None,
    initial_wtg_north_table: pd.DataFrame | None = None,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    wtg_df = wtg_df.copy()
    wtg_df = add_ok_yaw_col(
        wtg_df,
        new_col_name=f"ok_for_{north_ref_wd_col}_northing",
        wd_col=north_ref_wd_col,
        rated_power=rated_power,
        timebase_s=timebase_s,
    )
    wtg_df = add_northed_ok_diff_and_rolling_cols(
        wtg_df, north_ref_wd_col=north_ref_wd_col, timebase_s=timebase_s, north_offset=0
    )

    initial_score = northing_score(
        wtg_df,
        north_ref_wd_col=north_ref_wd_col,
        changepoint_count=0,
        rated_power=rated_power,
        timebase_s=timebase_s,
    )
    logger.info(f"\n\nwtg_name={wtg_name}, north_ref_wd_col={north_ref_wd_col}, initial_score={initial_score:.2f}")

    if plot_cfg is not None:
        plot_yaw_diff_vs_power(wtg_df, wtg_name=wtg_name, north_ref_wd_col=north_ref_wd_col, plot_cfg=plot_cfg)

    if initial_wtg_north_table is None or len(initial_wtg_north_table) == 0:
        wtg_north_table = pd.DataFrame(
            data={TIMESTAMP_COL: wtg_df.index.min(), "north_offset": 0.0},
            index=[0],
        )
    else:
        initial_wtg_north_table = clip_wtg_north_table(
            initial_wtg_north_table,
            wtg_df=wtg_df,
        )

        if not north_table_is_valid(initial_wtg_north_table, wtg_df=wtg_df, check_north_offset=True):
            msg = "initial_wtg_north_table is not valid"
            raise ValueError(msg)
        wtg_north_table = initial_wtg_north_table.copy()

    if not north_table_is_valid(wtg_north_table, wtg_df=wtg_df, check_north_offset=True):
        msg = "wtg_north_table is not valid"
        raise RuntimeError(msg)

    return wtg_df, wtg_north_table


def optimize_wtg_north_table(
    *,
    wtg_df: pd.DataFrame,
    wtg_name: str,
    rated_power: float,
    north_ref_wd_col: str,
    timebase_s: int,
    plot_cfg: PlotConfig | None,
    initial_wtg_north_table: pd.DataFrame | None = None,
    best_score_margin: float = 0,
) -> tuple[pd.DataFrame, pd.DataFrame, float, float]:
    wtg_df = wtg_df.copy()

    wtg_df, wtg_north_table = prep_for_optimize_wtg_north_table(
        wtg_df,
        wtg_name=wtg_name,
        north_ref_wd_col=north_ref_wd_col,
        rated_power=rated_power,
        timebase_s=timebase_s,
        plot_cfg=plot_cfg,
        initial_wtg_north_table=initial_wtg_north_table,
    )

    loop_count = 0
    best_north_table, initial_score, best_wtg_df = score_wtg_north_table(
        wtg_north_table=wtg_north_table,
        wtg_df=wtg_df,
        improve_north_offset_col=True,
        north_ref_wd_col=north_ref_wd_col,
        rated_power=rated_power,
        timebase_s=timebase_s,
    )
    best_score = initial_score
    logger.info(f"best_score={best_score:.2f} before optimization")

    if plot_cfg is not None:
        plot_diff_to_north_ref_wd(
            best_wtg_df,
            wtg_name=wtg_name,
            north_ref_wd_col=north_ref_wd_col,
            loop_count=loop_count,
            plot_cfg=plot_cfg,
        )

    done_optimizing = False
    max_changepoints_to_add = min(5, calc_max_changepoints_to_add(len(best_north_table), score=best_score))
    initial_step_size: int = 100
    shift_step_size: int = initial_step_size
    tries_left: int = 1
    while not done_optimizing:
        loop_count += 1
        best_move_found_this_loop = False
        prev_best_north_table = best_north_table.copy()
        prev_best_wtg_df = best_wtg_df.copy()

        do_changepoint_moves = max_changepoints_to_add > 0
        moves = list_possible_moves(
            prev_best_north_table,
            do_changepoint_moves=do_changepoint_moves,
            max_changepoints_to_add=max_changepoints_to_add,
        )
        if do_changepoint_moves:
            algo, timestamps = get_changepoint_objects(
                prev_best_wtg_df=prev_best_wtg_df,
                north_ref_wd_col=north_ref_wd_col,
            )

        for move in moves:
            this_north_table, this_score, this_wtg_df = make_move_and_score_wtg_north_table(
                move,
                prev_best_north_table=prev_best_north_table,
                wtg_df=wtg_df,
                shift_step_size=shift_step_size,
                north_ref_wd_col=north_ref_wd_col,
                rated_power=rated_power,
                timebase_s=timebase_s,
                do_changepoint_moves=do_changepoint_moves,
                algo=algo,
                timestamps=timestamps,
            )

            if this_score < (best_score - best_score_margin):
                best_north_table = this_north_table.copy()
                best_score = this_score
                best_wtg_df = this_wtg_df.copy()
                best_move_found_this_loop = True
                logger.info(
                    f"wtg_name={wtg_name}, best_score={this_score:.3f}, loop_count={loop_count}, "
                    f"shift_step_size={shift_step_size}, len(best_north_table)={len(best_north_table)}, "
                    f"move={move}",
                )
        if len(best_north_table) == len(prev_best_north_table):
            max_changepoints_to_add = 0
        else:
            tries_left += 1
            logger.info(f"tries_left increased to {tries_left}")
            max_changepoints_to_add = min(2, calc_max_changepoints_to_add(len(best_north_table), score=best_score))
        if not best_move_found_this_loop:
            shift_step_size = min(shift_step_size - 1, round(shift_step_size * DECAY_FRACTION))
            if shift_step_size < 1:
                max_changepoints_to_add = min(
                    2,
                    calc_max_changepoints_to_add(len(best_north_table), score=best_score),
                )
                shift_step_size = initial_step_size + 1 + (1 / (DECAY_FRACTION ** (math.pi * (tries_left + 1))) % 10)
                shift_step_size = round(shift_step_size)
                tries_left -= 1
                logger.info(f"tries_left decreased to {tries_left}")
        if tries_left == 0:
            done_optimizing = True

    wtg_north_table = best_north_table.copy()
    wtg_df = add_northed_ok_diff_and_rolling_cols(
        wtg_df,
        north_ref_wd_col=north_ref_wd_col,
        timebase_s=timebase_s,
        north_offset_df=wtg_north_table,
    )
    if plot_cfg is not None:
        plot_diff_to_north_ref_wd(
            wtg_df,
            wtg_name=wtg_name,
            north_ref_wd_col=north_ref_wd_col,
            loop_count=loop_count,
            plot_cfg=plot_cfg,
        )

    return wtg_north_table, wtg_df, initial_score, best_score


def optimize_wf_north_table(
    wf_df: pd.DataFrame,
    *,
    north_ref_wd_col: str,
    cfg: WindUpConfig,
    plot_cfg: PlotConfig | None,
    best_score_margin: float = 0,
) -> pd.DataFrame:
    optimized_wf_north_table = pd.DataFrame()
    initial_wf_north_table = pd.DataFrame(
        data=cfg.northing_corrections_utc,
        columns=["TurbineName", TIMESTAMP_COL, "north_offset"],
    )
    for wtg_name in sorted(wf_df.index.unique(level="TurbineName").to_list()):
        wtg_obj = next(x for x in cfg.asset.wtgs if x.name == wtg_name)
        rated_power = wtg_obj.turbine_type.rated_power_kw

        wtg_df = wf_df.loc[wtg_name].copy()
        max_northing_error_before = check_wtg_northing(
            wtg_df,
            wtg_name=wtg_name,
            north_ref_wd_col=north_ref_wd_col,
            timebase_s=cfg.timebase_s,
            plot_cfg=None,
        )

        initial_wtg_north_table = initial_wf_north_table.loc[initial_wf_north_table["TurbineName"] == wtg_name]
        changepoints_before = max(1, len(initial_wtg_north_table))
        wtg_north_table, optimized_wtg_df, score_before, score_after = optimize_wtg_north_table(
            wtg_df=wtg_df,
            wtg_name=wtg_name,
            rated_power=rated_power,
            north_ref_wd_col=north_ref_wd_col,
            timebase_s=cfg.timebase_s,
            plot_cfg=plot_cfg,
            initial_wtg_north_table=initial_wtg_north_table,
            best_score_margin=best_score_margin,
        )

        northed_col = calc_northed_col_name(north_ref_wd_col)
        optimized_wtg_df["YawAngleMean"] = optimized_wtg_df[northed_col]
        max_northing_error_after = check_wtg_northing(
            optimized_wtg_df,
            wtg_name=wtg_name,
            north_ref_wd_col=north_ref_wd_col,
            timebase_s=cfg.timebase_s,
            plot_cfg=plot_cfg,
        )

        changepoints_after = len(wtg_north_table)
        logger.info(
            f"changepoints changed from {changepoints_before} to {changepoints_after} "
            f"[{changepoints_after - changepoints_before}]",
        )
        logger.info(
            f"northing score changed from {score_before:.1f} to {score_after:.1f} [{score_after - score_before:.1f}]"
        )
        logger.info(
            f"max_northing_error changed from {max_northing_error_before:.1f} to {max_northing_error_after:.1f} "
            f"[{max_northing_error_after - max_northing_error_before:.1f}]",
        )

        wtg_north_table["TurbineName"] = wtg_name
        optimized_wf_north_table = (
            pd.concat([optimized_wf_north_table, wtg_north_table])
            .sort_values(by=["TurbineName", TIMESTAMP_COL])
            .reset_index(drop=True)
        )
    return optimized_wf_north_table


def write_northing_yaml(wf_north_table: pd.DataFrame, *, fpath: Path) -> None:
    north_table_for_yaml = wf_north_table.copy()
    north_table_for_yaml[TIMESTAMP_COL] = north_table_for_yaml[TIMESTAMP_COL].dt.strftime(
        "%Y-%m-%d %H:%M:%S",
    )
    yaml_strings = []
    for _, row in north_table_for_yaml.iterrows():
        yaml_strings.append(f"    - ['{row['TurbineName']}', {row[TIMESTAMP_COL]}, {row['north_offset']}]")
    yaml_content = "\n".join(yaml_strings)
    with fpath.open(mode="w") as yaml_file:
        yaml_file.write(yaml_content)


def auto_northing_corrections(
    wf_df: pd.DataFrame,
    *,
    cfg: WindUpConfig,
    plot_cfg: PlotConfig | None,
) -> pd.DataFrame:
    wf_df = wf_df.copy()

    reanalysis_wf_north_table = optimize_wf_north_table(
        wf_df,
        north_ref_wd_col=REANALYSIS_WD_COL,
        cfg=cfg,
        plot_cfg=plot_cfg,
        best_score_margin=0.5,
    )
    if plot_cfg is not None:
        reanalysis_wf_north_table.to_csv(cfg.out_dir / "reanalysis_wf_north_table.csv")
        write_northing_yaml(reanalysis_wf_north_table, fpath=cfg.out_dir / "reanalysis_wf_north_table.yaml")

    wf_df = apply_northing_corrections(
        wf_df,
        wf_north_table=reanalysis_wf_north_table,
        north_ref_wd_col=REANALYSIS_WD_COL,
        cfg=cfg,
        plot_cfg=plot_cfg,
    )

    wf_df = add_wf_yawdir(wf_df, cfg=cfg)

    if plot_cfg is not None:
        plot_wf_yawdir_and_reanalysis_timeseries(wf_df, cfg=cfg, plot_cfg=plot_cfg)

    optimized_northing_corrections = optimize_wf_north_table(
        wf_df,
        north_ref_wd_col=WINDFARM_YAWDIR_COL,
        cfg=cfg,
        plot_cfg=plot_cfg,
    )
    if plot_cfg is not None:
        optimized_northing_corrections.to_csv(cfg.out_dir / "optimized_northing_corrections.csv")
        write_northing_yaml(optimized_northing_corrections, fpath=cfg.out_dir / "optimized_northing_corrections.yaml")

    return apply_northing_corrections(
        wf_df,
        wf_north_table=optimized_northing_corrections,
        north_ref_wd_col=WINDFARM_YAWDIR_COL,
        cfg=cfg,
        plot_cfg=plot_cfg,
    )



import logging

import matplotlib.pyplot as plt
import pandas as pd
from tabulate import tabulate

from wind_up.constants import RAW_YAWDIR_COL
from wind_up.models import PlotConfig, WindUpConfig
from wind_up.northing_utils import add_ok_yaw_col

logger = logging.getLogger(__name__)


def plot_northing_error(
    wf_df: pd.DataFrame,
    title_end: str,
    plot_cfg: PlotConfig,
    sub_dir: str | None = None,
) -> None:
    title_end = f" {title_end}" if len(title_end) > 0 else ""
    plt.figure(figsize=(10, 6))
    wtg_names = wf_df.index.unique(level="TurbineName")
    for wtg_name in wtg_names:
        plt.plot(wf_df.loc[wtg_name, "rolling_northing_error"].resample("6h").mean(), label=wtg_name)
    plt.legend(loc="best", ncol=max(1, round(len(wtg_names) // 5)), fontsize="small")
    plot_title = f"northing error{title_end}"
    plt.title(plot_title)
    plt.xlabel("datetime")
    plt.ylabel("rolling_northing_error [deg]")
    plt.grid()
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        if sub_dir is None:
            plot_cfg.plots_dir.mkdir(exist_ok=True)
            plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
        else:
            (plot_cfg.plots_dir / sub_dir).mkdir(exist_ok=True)
            plt.savefig(plot_cfg.plots_dir / sub_dir / f"{plot_title}.png")
    plt.close()


def print_northing_error_summary(abs_north_errs: pd.DataFrame, *, wtgs_description: str, title_end: str) -> None:
    title_end = f" {title_end}" if len(title_end) > 0 else ""
    _table = tabulate((abs_north_errs.sort_values(ascending=False)[0:3]).to_frame(), tablefmt="outline", floatfmt=".1f")
    logger.info(f"top 3 {wtgs_description} needing northing correction{title_end}:\n{_table}")


def plot_and_print_northing_error(
    wf_df: pd.DataFrame,
    *,
    cfg: WindUpConfig,
    abs_north_errs: pd.DataFrame,
    title_end: str,
    plot_cfg: PlotConfig,
) -> None:
    print_northing_error_summary(abs_north_errs, wtgs_description="turbines", title_end=title_end)
    ref_abs_north_errs = abs_north_errs[[x.name for x in cfg.ref_wtgs]]
    print_northing_error_summary(ref_abs_north_errs, wtgs_description="REFERENCE turbines", title_end=title_end)

    plot_northing_error(
        wf_df=wf_df,
        title_end=title_end,
        plot_cfg=plot_cfg,
    )

    plot_northing_error(
        wf_df=wf_df.loc[[x.name for x in cfg.ref_wtgs]],
        title_end="REF wtgs " + title_end,
        plot_cfg=plot_cfg,
    )


def plot_northing_changepoint(
    wf_df: pd.DataFrame,
    *,
    northing_turbine: str,
    northed_col: str,
    north_ref_wd_col: str,
    northing_datetime_utc: pd.Timestamp,
    cfg: WindUpConfig,
    plot_cfg: PlotConfig,
) -> None:
    cols_to_plot = [RAW_YAWDIR_COL, northed_col, north_ref_wd_col]
    plot_df = wf_df.loc[northing_turbine].copy()
    days_to_plot = 2
    xmin = northing_datetime_utc - pd.Timedelta(days=days_to_plot)
    xmax = northing_datetime_utc + pd.Timedelta(days=days_to_plot)
    plot_df = plot_df.loc[xmin:xmax, :].dropna(subset=cols_to_plot, how="all")

    plt.figure()
    plt.plot(plot_df.index, plot_df[north_ref_wd_col], label=north_ref_wd_col)
    plt.plot(plot_df.index, plot_df[RAW_YAWDIR_COL], label=RAW_YAWDIR_COL)
    plt.plot(
        plot_df.index,
        plot_df[northed_col],
        label=northed_col,
    )
    plot_df = add_ok_yaw_col(
        plot_df,
        new_col_name="ok_for_yawdir",
        wd_col=northed_col,
        rated_power=next(x for x in cfg.asset.wtgs if x.name == northing_turbine).turbine_type.rated_power_kw,
        timebase_s=cfg.timebase_s,
    )
    plot_df = plot_df.loc[plot_df["ok_for_yawdir"]]
    plt.plot(
        plot_df.index,
        plot_df[northed_col],
        marker=".",
        linestyle="None",
        label=f"FILTERED yawdir_northed_to_{north_ref_wd_col}",
        color="darkgreen",
    )
    plt.axvline(northing_datetime_utc, color="grey", linestyle="--", label="northing correction")
    plt.xlabel("timestamp")
    plt.ylabel("direction [deg]")
    plt.xlim([xmin, xmax])
    plt.xticks(rotation=45)
    title = f"{northing_turbine} north_ref_wd_col={north_ref_wd_col} {northing_datetime_utc.strftime('%Y-%m-%d')}"
    plt.title(title)
    plt.legend(fontsize="small", ncol=2)
    plt.grid()
    plt.tight_layout()
    (plot_cfg.plots_dir / northing_turbine).mkdir(exist_ok=True)
    plt.savefig(plot_cfg.plots_dir / northing_turbine / f"{title}.png")
    plt.close()



import logging
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import utm

from wind_up.models import WindUpConfig

logger = logging.getLogger(__name__)


def rescale_vector(
    vector: pd.Series | np.ndarray,
    lowthres: float = 0,
    highthres: float = 1,
) -> pd.Series | np.ndarray:
    if isinstance(vector, list):
        vector = np.array(vector)

    delta = highthres - lowthres
    tozero = vector - vector.min()
    if not (tozero != 0).any():
        tozero += 1
    zeroone = tozero / (tozero.max())
    scaled = zeroone * delta
    return scaled + lowthres


def bubble_plot_calcs(
    *,
    cfg: WindUpConfig,
    series: pd.Series,
    sizes: pd.Series,
    cbarunits: str | None = None,
) -> tuple[
    np.ndarray,
    np.ndarray,
    tuple[float, float],
    tuple[float, float],
    np.ndarray,
    np.ndarray,
    np.ndarray,
    list[str],
    list[str],
    list[str],
    list[str],
]:
    x = np.array([])
    y = np.array([])
    wtg_id = []
    for wtg in cfg.asset.wtgs:
        c = utm.from_latlon(wtg.latitude, wtg.longitude)
        x = np.append(x, c[0])
        y = np.append(y, c[1])
        wtg_id.append(wtg.name)
    x = x - np.median(x)
    y = y - np.median(y)

    xmar = 500
    ymar = 500
    xlims = (x.min() - xmar, x.max() + xmar)
    ylims = (y.min() - ymar, y.max() + ymar)

    cbarstr = "" if cbarunits is None else cbarunits

    # negative numbers in the series
    if any(series < 0):
        # color
        logger.warning("any(series<0)=True, perhaps some the sizes are weird")
        s = series + abs(series.min())
        sc = 0.1 + 0.8 * s / s.max()
        midtick = (-series.min() * 0.8) / (series.max() - series.min()) + 0.1
        ticks = np.array([sc.min(), midtick, sc.max()])
        ticklabs = [
            f"{series.dropna().min():3.1f}\n{cbarstr}",
            f"{0:3.1f}\n{cbarstr}",
            f"{series.dropna().max():3.1f}\n{cbarstr}",
        ]
        # size
        si = sizes.abs() / sizes.abs().max()

    # all positive numbers
    else:
        # color
        s = rescale_vector(series, lowthres=0.01, highthres=0.99)
        sc = s.copy()
        ticks = np.linspace(sc.min(), sc.max(), 3)
        ticklabs = [
            f"{series.min():3.1f}\n{cbarstr}",
            f"{(series.max() + series.min()) * 0.5:3.1f}\n{cbarstr}",
            f"{series.max():3.1f}\n{cbarstr}",
        ]
        # size
        si = rescale_vector(sizes, lowthres=0, highthres=2)

    park_wtgids = [wtg.name for wtg in cfg.asset.wtgs]
    series_wtgs = series.index.tolist()
    missing_wtgs = [wtg.name for wtg in cfg.asset.wtgs if wtg.name not in series_wtgs]

    if missing_wtgs:
        misseries = pd.Series(index=missing_wtgs)
        sc = pd.concat([sc, misseries], axis=0, sort=True)
        sc = sc.loc[park_wtgids].copy()  # right order
        si = pd.concat([si, misseries], axis=0, sort=True)
        si = si.loc[park_wtgids].copy()  # right order

    if len(sc) != len(park_wtgids):
        msg = "The length of 'series' does not match with the number of turbines"
        raise ValueError(msg)

    if len(si) != len(park_wtgids):
        msg = "The length of 'sizes' does not match with the number of turbines"
        raise ValueError(msg)

    return x, y, xlims, ylims, sc, si, ticks, ticklabs, wtg_id, park_wtgids, series_wtgs


def bubble_plot(
    cfg: WindUpConfig,
    series: pd.Series,
    *,
    sizes: pd.Series | None = None,
    cmap: str = "coolwarm",
    fontsize: int = 12,
    idfontsize: int = 10,
    bubblesize: int = 1000,
    movetext: tuple[float, float] = (-0.0017, -0.001),
    txt: bool = True,
    title: str | None = None,
    cbarunits: str | None = None,
    save_path: Path | None = None,
    display_wtgids: bool = True,
    cbar_label: str | None = None,
    text_rotation: int = 0,
    figuresize: tuple[float, float] | None = None,
    show_plot: bool = False,
) -> None:
    if isinstance(series, type(pd.DataFrame())):
        msg = "series should be a pandas series"
        raise TypeError(msg)

    if sizes is None:
        sizes = series

    x, y, xlims, ylims, sc, si, ticks, ticklabs, wtg_id, park_wtgids, series_wtgs = bubble_plot_calcs(
        cfg=cfg,
        series=series,
        sizes=sizes,
        cbarunits=cbarunits,
    )

    # meters will have the same dimension in both axes
    ratio = (ylims[1] - ylims[0]) / (xlims[1] - xlims[0])
    hsize = 10
    if figuresize is None:
        f, ax = plt.subplots(1, 1, figsize=(hsize * 1.2, hsize * ratio))
    else:
        f, ax = plt.subplots(1, 1, figsize=figuresize)

    csa = ax.scatter(x, y, c=sc, s=bubblesize * si + 100, cmap=cmap, edgecolors=None)

    cbar = f.colorbar(csa, ticks=ticks, pad=0.02, fraction=0.046)  # cax=cax)
    if cbar_label is not None:
        cbar.set_label(cbar_label, fontsize=fontsize)
    cbar.ax.set_yticklabels(ticklabs, fontsize=fontsize)

    plt.xlim(xlims)
    plt.ylim(ylims)
    plt.tick_params(axis="x", labelsize=fontsize)
    plt.tick_params(axis="y", labelsize=fontsize)

    plt.ylabel("Y coord [m]", fontsize=fontsize)
    plt.xlabel("X coord [m]", fontsize=fontsize)
    if title is None:
        plt.title(cfg.asset.name, fontsize=int(fontsize * 1.1))
    else:
        plt.title(title, fontsize=int(fontsize * 1.1))
    plt.tight_layout()

    if txt:
        for i in range(len(park_wtgids)):
            if display_wtgids:
                if wtg_id[i] in series_wtgs:
                    plt.text(
                        x[i] + movetext[0],
                        y[i] + movetext[1],
                        wtg_id[i],
                        color="k",
                        weight=600,
                        fontsize=idfontsize,
                        rotation=text_rotation,
                    )
                else:
                    plt.text(
                        x[i] + movetext[0],
                        y[i] + movetext[1],
                        wtg_id[i],
                        color="grey",
                        fontsize=idfontsize,
                        rotation=text_rotation,
                    )

    plt.grid()
    if show_plot:
        plt.show()
    if save_path is not None:
        plt.savefig(save_path, dpi=120)
    plt.close()



import logging

import numpy as np
import pandas as pd

from wind_up.constants import HOURS_PER_YEAR, RAW_POWER_COL, RAW_WINDSPEED_COL
from wind_up.models import PlotConfig, WindUpConfig
from wind_up.plots.long_term_plots import plot_lt_ws, plot_lt_ws_raw_filt
from wind_up.result_manager import result_manager

logger = logging.getLogger(__name__)


def calc_lt_df(
    df_for_lt: pd.DataFrame,
    *,
    num_turbines: int,
    years_for_lt_distribution: int,
    ws_col: str,
    ws_bin_width: float,
    pw_col: str,
    timebase_s: int,
) -> pd.DataFrame:
    years_of_data = (df_for_lt.index.max() - df_for_lt.index.min()).total_seconds() / 3600 / 24 / 365.25
    if years_of_data < 11.5 / 12:
        result_manager.warning(f"years_of_data for long term is too small: {years_of_data:.1f}")
    max_years_error = 0.1
    if years_for_lt_distribution - years_of_data > max_years_error:
        result_manager.warning(
            "years_of_data for long term is small, "
            f"expected {years_for_lt_distribution:.1f} actual {years_of_data:.1f}",
        )

    if not isinstance(df_for_lt.index, pd.DatetimeIndex):
        msg = f"df_for_lt must have a DatetimeIndex, got {type(df_for_lt.index)}"
        raise TypeError(msg)
    counts_by_month = df_for_lt.index.month.value_counts().to_frame()
    if len(counts_by_month) < 12:  # noqa PLR2004
        result_manager.warning(f"only {len(counts_by_month)} months represented in data for long term")
    counts_by_hour = df_for_lt.index.hour.value_counts().to_frame()
    if len(counts_by_hour) < 24:  # noqa PLR2004
        result_manager.warning(f"only {len(counts_by_hour)} hours represented in data for long term")

    ws_bin_edges = np.arange(0, df_for_lt[ws_col].max() + ws_bin_width, ws_bin_width)

    rows_per_hour = 3600 / timebase_s
    lt_df = (
        df_for_lt.dropna(subset=[ws_col, pw_col])
        .groupby(by=pd.cut(df_for_lt[ws_col], bins=ws_bin_edges, retbins=False), observed=False)
        .agg(
            ws_mean=pd.NamedAgg(column=ws_col, aggfunc=lambda x: x.mean()),
            observed_hours=pd.NamedAgg(column=ws_col, aggfunc=lambda x: len(x) / rows_per_hour),
            observed_mwh=pd.NamedAgg(column=pw_col, aggfunc=lambda x: sum(x) / rows_per_hour / 1000),
        )
    )
    lt_df = lt_df.rename(columns={"ws_mean": f"{ws_col}_mean"})
    lt_df["bin_left"] = [x.left for x in lt_df.index]
    lt_df["bin_mid"] = [x.mid for x in lt_df.index]
    lt_df["bin_right"] = [x.right for x in lt_df.index]
    lt_df["bin_closed_right"] = [x.closed_right for x in lt_df.index]

    lt_df = lt_df.set_index("bin_mid", drop=False, verify_integrity=True)
    lt_df.index.name = f"{ws_col}_bin_mid"

    lt_df[f"{ws_col}_mean"] = lt_df[f"{ws_col}_mean"].fillna(lt_df["bin_mid"])

    lt_df["fraction_of_time"] = lt_df["observed_hours"] / lt_df["observed_hours"].sum()
    if years_for_lt_distribution > 0:
        data_coverage = lt_df["observed_hours"].sum() / (years_for_lt_distribution * HOURS_PER_YEAR) / num_turbines
    else:
        data_coverage = lt_df["observed_hours"].sum() / (years_of_data * HOURS_PER_YEAR) / num_turbines
        result_manager.warning("years_for_lt_distribution is 0, using years_of_data instead to calculate data coverage")
    if data_coverage > 1:
        msg = f"Data coverage for long term ws distribution is >1, {data_coverage:.2f}"
        raise RuntimeError(msg)
    lt_df["hours_per_year"] = lt_df["fraction_of_time"] * HOURS_PER_YEAR * data_coverage
    lt_df["mwh_per_year_per_turbine"] = (
        lt_df["hours_per_year"] * lt_df["observed_mwh"] / lt_df["observed_hours"]
    ).fillna(0)

    if lt_df.isna().any().any():
        msg = "lt_df has missing values"
        raise RuntimeError(msg)

    logger.info(f"long term distribution uses data from {df_for_lt.index.min()} to {df_for_lt.index.max()}")
    logger.info(f"long term distribution data coverage: {data_coverage*100:.1f}%")

    return lt_df


def calc_turbine_lt_df(
    wtg_name: str,
    cfg: WindUpConfig,
    wtg_df: pd.DataFrame,
    *,
    ws_col: str,
    pw_col: str,
    title_end: str = "",
    plot_cfg: PlotConfig | None = None,
) -> pd.DataFrame:
    workings_df = wtg_df.copy()
    if not isinstance(workings_df.index, pd.DatetimeIndex):
        msg = f"wtg_df must have a DatetimeIndex, got {type(workings_df.index)}"
        raise TypeError(msg)

    ok_for_lt = (workings_df.index >= cfg.lt_first_dt_utc_start) & (workings_df.index <= cfg.lt_last_dt_utc_start)

    lt_df = calc_lt_df(
        df_for_lt=workings_df[ok_for_lt],
        num_turbines=1,
        years_for_lt_distribution=cfg.years_for_lt_distribution,
        ws_col=ws_col,
        ws_bin_width=cfg.ws_bin_width,
        pw_col=pw_col,
        timebase_s=cfg.timebase_s,
    )
    if plot_cfg is not None:
        plot_lt_ws(lt_df=lt_df, turbine_or_wf_name=wtg_name, title_end=title_end, plot_cfg=plot_cfg, one_turbine=True)

    return lt_df


def calc_turbine_lt_dfs_raw_filt(
    wtg_name: str,
    cfg: WindUpConfig,
    wtg_df: pd.DataFrame,
    *,
    ws_col: str,
    pw_col: str,
    plot_cfg: PlotConfig | None = None,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    lt_wf_df_raw = calc_turbine_lt_df(
        wtg_name=wtg_name,
        cfg=cfg,
        wtg_df=wtg_df,
        ws_col=RAW_WINDSPEED_COL,
        pw_col=RAW_POWER_COL,
        title_end="before filter",
        plot_cfg=plot_cfg,
    )
    lt_wf_df_filt = calc_turbine_lt_df(
        wtg_name=wtg_name,
        cfg=cfg,
        wtg_df=wtg_df,
        ws_col=ws_col,
        pw_col=pw_col,
        title_end="after filter",
        plot_cfg=plot_cfg,
    )
    if plot_cfg is not None:
        plot_lt_ws_raw_filt(
            lt_df_raw=lt_wf_df_raw,
            lt_df_filt=lt_wf_df_filt,
            turbine_or_wf_name=wtg_name,
            plot_cfg=plot_cfg,
            one_turbine=True,
        )

    return lt_wf_df_raw, lt_wf_df_filt


def calc_windfarm_lt_df(
    cfg: WindUpConfig,
    wf_df: pd.DataFrame,
    *,
    ws_col: str,
    pw_col: str,
    title_end: str = "",
    plot_cfg: PlotConfig | None = None,
) -> pd.DataFrame:
    workings_df = wf_df.copy()
    if len(workings_df.index.levels) == 2:  # noqa PLR2004
        workings_df.index = workings_df.index.droplevel("TurbineName")

    ok_for_lt = (workings_df.index >= cfg.lt_first_dt_utc_start) & (workings_df.index <= cfg.lt_last_dt_utc_start)

    lt_df = calc_lt_df(
        df_for_lt=workings_df[ok_for_lt],
        num_turbines=len(cfg.asset.wtgs),
        years_for_lt_distribution=cfg.years_for_lt_distribution,
        ws_col=ws_col,
        ws_bin_width=cfg.ws_bin_width,
        pw_col=pw_col,
        timebase_s=cfg.timebase_s,
    )

    if plot_cfg is not None:
        plot_lt_ws(lt_df=lt_df, turbine_or_wf_name=cfg.asset.name, title_end=title_end, plot_cfg=plot_cfg)

    return lt_df


def calc_windfarm_lt_dfs_raw_filt(
    cfg: WindUpConfig,
    wf_df_raw: pd.DataFrame,
    wf_df_filt: pd.DataFrame,
    *,
    ws_col: str,
    pw_col: str,
    plot_cfg: PlotConfig | None = None,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    lt_wf_df_raw = calc_windfarm_lt_df(
        cfg=cfg,
        wf_df=wf_df_raw,
        ws_col=ws_col,
        pw_col=pw_col,
        title_end="before filter",
        plot_cfg=plot_cfg,
    )
    lt_wf_df_filt = calc_windfarm_lt_df(
        cfg=cfg,
        wf_df=wf_df_filt,
        ws_col=ws_col,
        pw_col=pw_col,
        title_end="after filter",
        plot_cfg=plot_cfg,
    )
    if plot_cfg is not None:
        plot_lt_ws_raw_filt(
            lt_df_raw=lt_wf_df_raw,
            lt_df_filt=lt_wf_df_filt,
            turbine_or_wf_name=cfg.asset.name,
            plot_cfg=plot_cfg,
        )

    return lt_wf_df_raw, lt_wf_df_filt



import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from tabulate import tabulate

from wind_up.constants import SCATTER_ALPHA, SCATTER_MARKERSCALE, SCATTER_S, DataColumns
from wind_up.models import PlotConfig, WindUpConfig
from wind_up.plots.misc_plots import bubble_plot

logger = logging.getLogger(__name__)


def plot_data_coverage_heatmap(df: pd.DataFrame, plot_title: str, plot_cfg: PlotConfig) -> None:
    # calculate data coverage per turbine
    covdf = df.groupby("TurbineName", observed=False).agg(
        power=pd.NamedAgg(column="ActivePowerMean", aggfunc=lambda x: x.count() / x.size),
        windspeed=pd.NamedAgg(column="WindSpeedMean", aggfunc=lambda x: x.count() / x.size),
        yaw=pd.NamedAgg(column="YawAngleMean", aggfunc=lambda x: x.count() / x.size),
        rpm=pd.NamedAgg(column="GenRpmMean", aggfunc=lambda x: x.count() / x.size),
        pitch=pd.NamedAgg(column="PitchAngleMean", aggfunc=lambda x: x.count() / x.size),
    )

    plt.figure()
    sns.heatmap(covdf, annot=True, fmt=".2f", vmax=1, vmin=min(0.5, covdf.min().min()))
    plt.title(plot_title)
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / f"{plot_title}.png")
    if plot_cfg.show_plots:
        plt.show()
    plt.close()


def calc_cf_by_turbine(scada_df: pd.DataFrame, cfg: WindUpConfig) -> pd.DataFrame:
    rows_per_hour = 3600 / cfg.timebase_s
    cf_df = scada_df.groupby("TurbineName", observed=False).agg(
        hours=pd.NamedAgg(column="TurbineName", aggfunc=lambda x: x.count() / rows_per_hour),
        MWh=pd.NamedAgg(column="ActivePowerMean", aggfunc=lambda x: x.sum() / rows_per_hour / 1000),
    )
    for i, rp in zip(
        [x.name for x in cfg.asset.wtgs],
        [x.turbine_type.rated_power_kw for x in cfg.asset.wtgs],
        strict=True,
    ):
        cf_df.loc[i, "rated_power_kW"] = rp
    cf_df["CF"] = cf_df["MWh"] / (cf_df["hours"] * cf_df["rated_power_kW"] / 1000)
    return cf_df


def print_and_plot_capacity_factor(scada_df: pd.DataFrame, cfg: WindUpConfig, plots_cfg: PlotConfig) -> None:
    cf_df = calc_cf_by_turbine(scada_df=scada_df, cfg=cfg)
    title = f"{cfg.asset.name} capacity factor"
    plots_cfg.plots_dir.mkdir(parents=True, exist_ok=True)
    bubble_plot(
        cfg=cfg,
        series=cf_df["CF"] * 100,
        title=f"{cfg.asset.name} capacity factor",
        cbarunits="%",
        save_path=plots_cfg.plots_dir / f"{title}.png",
        show_plot=plots_cfg.show_plots,
    )

    logger.info(f'average capacity factor: {cf_df["CF"].mean() * 100:.1f}%')
    _table = tabulate(
        (cf_df.sort_values(by="CF", ascending=False)["CF"][0:3] * 100).to_frame(),
        tablefmt="outline",
        floatfmt=".1f",
    )
    logger.info(f"top 3 capacity factor [%]:\n{_table}")
    _table = tabulate((cf_df.sort_values(by="CF")["CF"][0:3] * 100).to_frame(), tablefmt="outline", floatfmt=".1f")
    logger.info(f"bottom 3 capacity factor [%]:\n{_table}")


def plot_ops_curves_per_ttype(cfg: WindUpConfig, df: pd.DataFrame, title_end: str, plot_cfg: PlotConfig) -> None:
    for ttype in cfg.list_unique_turbine_types():
        wtgs = cfg.list_turbine_ids_of_type(ttype)
        df_ttype = df.loc[wtgs]
        plot_ops_curves_one_ttype_or_wtg(
            df=df_ttype,
            ttype_or_wtg=ttype.turbine_type,
            title_end=title_end,
            plot_cfg=plot_cfg,
        )
        if not plot_cfg.skip_per_turbine_plots:
            for wtg in wtgs:
                plot_ops_curves_one_ttype_or_wtg(
                    df=df_ttype.loc[[wtg]],
                    ttype_or_wtg=wtg,
                    title_end=title_end,
                    plot_cfg=plot_cfg,
                )
                plot_ops_curve_timelines_one_wtg(
                    wtg_df=df_ttype.loc[wtg],
                    wtg_name=wtg,
                    title_end=title_end,
                    plot_cfg=plot_cfg,
                )


def plot_ops_curves_one_ttype_or_wtg(df: pd.DataFrame, ttype_or_wtg: str, title_end: str, plot_cfg: PlotConfig) -> None:
    plt.figure()
    plt.scatter(df["WindSpeedMean"], df["ActivePowerMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plot_title = f"{ttype_or_wtg} power curve {title_end}"
    plt.title(plot_title)
    plt.xlabel("WindSpeedMean [m/s]")
    plt.ylabel("ActivePowerMean [kW]")
    plt.grid()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        t_dir = plot_cfg.plots_dir / ttype_or_wtg
        t_dir.mkdir(exist_ok=True, parents=True)
        plt.savefig(t_dir / f"{plot_title}.png")
    plt.close()

    # plot rpm and pitch vs power and wind speed in a 2 by 2 grid
    plt.figure(figsize=(12, 8))
    plt.subplot(2, 2, 1)
    plt.scatter(df["ActivePowerMean"], df["GenRpmMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.xlabel("ActivePowerMean [kW]")
    plt.ylabel("GenRpmMean [RPM]")
    plt.grid()

    plt.subplot(2, 2, 2)
    plt.scatter(df["WindSpeedMean"], df["GenRpmMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.xlabel("WindSpeedMean [m/s]")
    plt.ylabel("GenRpmMean [RPM]")
    plt.grid()

    plt.subplot(2, 2, 3)
    plt.scatter(df["ActivePowerMean"], df["PitchAngleMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.xlabel("ActivePowerMean [kW]")
    plt.ylabel("PitchAngleMean [deg]")
    plt.grid()

    plt.subplot(2, 2, 4)
    plt.scatter(df["WindSpeedMean"], df["PitchAngleMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.xlabel("WindSpeedMean [m/s]")
    plt.ylabel("PitchAngleMean [deg]")
    plt.grid()

    plot_title = f"{ttype_or_wtg} ops curves, {title_end}"
    plt.suptitle(plot_title)
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(t_dir / f"{plot_title}.png")
    plt.close()


def plot_ops_curve_timelines_one_wtg(wtg_df: pd.DataFrame, wtg_name: str, title_end: str, plot_cfg: PlotConfig) -> None:
    dropna_df = wtg_df.dropna(
        subset=[
            DataColumns.wind_speed_mean,
            DataColumns.active_power_mean,
            DataColumns.gen_rpm_mean,
            DataColumns.pitch_angle_mean,
        ]
    )
    gen_df = dropna_df[dropna_df[DataColumns.active_power_mean] > 0].copy()

    for descr, x_var, y_var, x_bin_width in [
        ("power curve", DataColumns.wind_speed_mean, DataColumns.active_power_mean, 1),
        ("rpm v power", DataColumns.active_power_mean, DataColumns.gen_rpm_mean, 0),
        ("pitch v ws", DataColumns.wind_speed_mean, DataColumns.pitch_angle_mean, 1),
    ]:
        bins = np.arange(0, gen_df[x_var].max() + x_bin_width, x_bin_width) if x_bin_width > 0 else 10
        mean_curve = gen_df.groupby(pd.cut(gen_df[x_var], bins=bins, retbins=False), observed=True).agg(
            x_mean=pd.NamedAgg(column=x_var, aggfunc="mean"),
            y_mean=pd.NamedAgg(column=y_var, aggfunc="mean"),
        )
        gen_df["expected_y"] = np.interp(gen_df[x_var], mean_curve["x_mean"], mean_curve["y_mean"])

        daily_df = gen_df.resample("D").mean()
        monthly_df = gen_df.resample("ME").mean()
        if y_var == DataColumns.pitch_angle_mean:
            daily_df["relative_y"] = daily_df[y_var] - daily_df["expected_y"]
            monthly_df["relative_y"] = monthly_df[y_var] - monthly_df["expected_y"]
        else:
            daily_df["relative_y"] = (daily_df[y_var] / daily_df["expected_y"]).clip(0.5, 1.5)
            monthly_df["relative_y"] = (monthly_df[y_var] / monthly_df["expected_y"]).clip(0.5, 1.5)

        plt.figure()
        plt.plot(daily_df.index, daily_df["relative_y"], label="daily")
        plt.plot(monthly_df.index, monthly_df["relative_y"], label="monthly")
        plot_title = f"{wtg_name} relative {descr} timeline {title_end}"
        plt.title(plot_title)
        plt.xlabel("date")
        plt.ylabel(f"relative {descr}")
        plt.legend()
        plt.grid()
        plt.tight_layout()
        if plot_cfg.show_plots:
            plt.show()
        if plot_cfg.save_plots:
            t_dir = plot_cfg.plots_dir / wtg_name
            plt.savefig(t_dir / f"{plot_title}.png")
        plt.close()


def plot_toggle_ops_curves_one_ttype_or_wtg(
    input_df: pd.DataFrame,
    ttype_or_wtg: str,
    title_end: str,
    toggle_name: str,
    plot_cfg: PlotConfig,
    ws_col: str = "test_WindSpeedMean",
    pw_col: str = "test_ActivePowerMean",
    pt_col: str = "test_PitchAngleMean",
    rpm_col: str = "test_GenRpmMean",
) -> None:
    pd.set_option("future.no_silent_downcasting", True)  # noqa FBT003
    if "toggle_on" not in input_df.columns or "toggle_off" not in input_df.columns:
        df_off = input_df[input_df["test_toggle_off"].fillna(value=False).infer_objects(copy=False)]
        df_on = input_df[input_df["test_toggle_on"].fillna(value=False).infer_objects(copy=False)]
    else:
        df_off = input_df[input_df["toggle_off"].fillna(value=False).infer_objects(copy=False)]
        df_on = input_df[input_df["toggle_on"].fillna(value=False).infer_objects(copy=False)]

    plt.figure()
    plt.scatter(
        df_off[ws_col],
        df_off[pw_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"{toggle_name} OFF",
    )
    plt.scatter(
        df_on[ws_col],
        df_on[pw_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"{toggle_name} ON",
    )
    plot_title = f"{ttype_or_wtg} power curve by {toggle_name}, {title_end}"
    plt.title(plot_title)
    plt.xlabel(f"{ws_col} [m/s]")
    plt.ylabel(f"{pw_col} [kW]")
    plt.grid()
    plt.legend(loc="best", markerscale=SCATTER_MARKERSCALE)
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        t_dir = plot_cfg.plots_dir / ttype_or_wtg
        t_dir.mkdir(exist_ok=True, parents=True)
        plt.savefig(t_dir / f"{plot_title}.png")
    plt.close()

    # plot rpm and pitch vs power and wind speed in a 2 by 2 grid
    plt.figure(figsize=(12, 8))
    plt.subplot(2, 2, 1)
    plt.scatter(
        df_off[pw_col],
        df_off[rpm_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"{toggle_name} OFF",
    )
    plt.scatter(
        df_on[pw_col],
        df_on[rpm_col],
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
        label=f"{toggle_name} ON",
    )
    plt.xlabel(f"{pw_col} [kW]")
    plt.ylabel(f"{rpm_col} [RPM]")
    plt.grid()
    plt.legend(loc="best", markerscale=SCATTER_MARKERSCALE)

    plt.subplot(2, 2, 2)
    plt.scatter(df_off[ws_col], df_off[rpm_col], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.scatter(df_on[ws_col], df_on[rpm_col], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.xlabel(f"{ws_col} [m/s]")
    plt.ylabel(f"{rpm_col} [RPM]")
    plt.grid()

    plt.subplot(2, 2, 3)
    plt.scatter(df_off[pw_col], df_off[pt_col], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.scatter(df_on[pw_col], df_on[pt_col], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.xlabel(f"{pw_col} [kW]")
    plt.ylabel(f"{pt_col} [deg]")
    plt.grid()

    plt.subplot(2, 2, 4)
    plt.scatter(df_off[ws_col], df_off[pt_col], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.scatter(df_on[ws_col], df_on[pt_col], s=SCATTER_S, alpha=SCATTER_ALPHA)
    plt.xlabel(f"{ws_col} [m/s]")
    plt.ylabel(f"{pt_col} [deg]")
    plt.grid()

    plot_title = f"{ttype_or_wtg} ops curves by {toggle_name}, {title_end}"
    plt.suptitle(plot_title)
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(t_dir / f"{plot_title}.png")
    plt.close()


def compare_ops_curves_pre_post(
    pre_df: pd.DataFrame,
    post_df: pd.DataFrame,
    *,
    test_name: str,
    plot_cfg: PlotConfig,
    is_toggle_test: bool,
) -> None:
    if is_toggle_test:
        plot_toggle_ops_curves_one_ttype_or_wtg(
            input_df=pd.concat([pre_df, post_df]),
            ttype_or_wtg=test_name,
            title_end="power performance data",
            toggle_name="toggle",
            plot_cfg=plot_cfg,
        )
    else:
        pre_df_fake_toggle = pre_df.copy()
        post_df_fake_toggle = post_df.copy()
        pre_df_fake_toggle["test_toggle_off"] = True
        post_df_fake_toggle["test_toggle_on"] = True
        plot_toggle_ops_curves_one_ttype_or_wtg(
            input_df=pd.concat([pre_df_fake_toggle, post_df_fake_toggle]),
            ttype_or_wtg=test_name,
            title_end="power performance data",
            toggle_name="upgrade",
            plot_cfg=plot_cfg,
        )


def plot_filter_rpm_and_pt_curve_one_ttype_or_wtg(
    df: pd.DataFrame,
    ttype_or_wtg: str,
    pt_v_pw_curve: pd.DataFrame,
    pt_v_ws_curve: pd.DataFrame,
    rpm_v_pw_curve: pd.DataFrame,
    rpm_v_ws_curve: pd.DataFrame,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure(figsize=(12, 8))
    plt.subplot(2, 2, 1)
    plt.scatter(df["pw_clipped"], df["GenRpmMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    x = [rpm_v_pw_curve.index[0].left] + [x.mid for x in rpm_v_pw_curve.index] + [rpm_v_pw_curve.index[-1].right]
    y = [rpm_v_pw_curve["y_limit"].iloc[0], *list(rpm_v_pw_curve["y_limit"]), rpm_v_pw_curve["y_limit"].iloc[-1]]
    plt.plot(x, y, color="red")
    plt.xlabel("pw_clipped [kW]")
    plt.ylabel("GenRpmMean [deg]")
    plt.grid()

    plt.subplot(2, 2, 2)
    plt.scatter(df["WindSpeedMean"], df["GenRpmMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    x = [rpm_v_ws_curve.index[0].left] + [x.mid for x in rpm_v_ws_curve.index] + [rpm_v_ws_curve.index[-1].right]
    y = [rpm_v_ws_curve["y_limit"].iloc[0], *list(rpm_v_ws_curve["y_limit"]), rpm_v_ws_curve["y_limit"].iloc[-1]]
    plt.plot(x, y, color="red")
    plt.xlabel("WindSpeedMean [m/s]")
    plt.ylabel("GenRpmMean [deg]")
    plt.grid()

    plt.subplot(2, 2, 3)
    plt.scatter(df["pw_clipped"], df["PitchAngleMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    x = [pt_v_pw_curve.index[0].left] + [x.mid for x in pt_v_pw_curve.index] + [pt_v_pw_curve.index[-1].right]
    y = [pt_v_pw_curve["y_limit"].iloc[0], *list(pt_v_pw_curve["y_limit"]), pt_v_pw_curve["y_limit"].iloc[-1]]
    plt.plot(x, y, color="red")
    plt.xlabel("pw_clipped [kW]")
    plt.ylabel("PitchAngleMean [deg]")
    plt.grid()

    plt.subplot(2, 2, 4)
    plt.scatter(df["WindSpeedMean"], df["PitchAngleMean"], s=SCATTER_S, alpha=SCATTER_ALPHA)
    x = [pt_v_ws_curve.index[0].left] + [x.mid for x in pt_v_ws_curve.index] + [pt_v_ws_curve.index[-1].right]
    y = [pt_v_ws_curve["y_limit"].iloc[0], *list(pt_v_ws_curve["y_limit"]), pt_v_ws_curve["y_limit"].iloc[-1]]
    plt.plot(x, y, color="red")
    plt.xlabel("WindSpeedMean [m/s]")
    plt.ylabel("PitchAngleMean [deg]")
    plt.grid()

    plot_title = f"{ttype_or_wtg} rpm and pitch curve filters"
    plt.suptitle(plot_title)
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        (plot_cfg.plots_dir / ttype_or_wtg).mkdir(exist_ok=True, parents=True)
        plt.savefig(plot_cfg.plots_dir / ttype_or_wtg / f"{plot_title}.png")
    plt.close()


def print_filter_stats(
    filter_name: str,
    na_rows: int,
    total_rows: int,
    *,
    just_yaw: bool = False,
    just_min_max: bool = False,
    reason: str = "",
) -> None:
    min_max_str = " Min & Max" if just_min_max else ""
    if len(reason) > 0:
        reason = f" because of {reason}"
    if just_yaw:
        logger.info(
            f"{filter_name} set {na_rows} rows [{100 * na_rows / total_rows:.1f}%] to NA yaw{min_max_str}{reason}"
        )
    else:
        logger.info(f"{filter_name} set {na_rows} rows [{100 * na_rows / total_rows:.1f}%] to NA{reason}")



from pathlib import Path

import pandas as pd
import pytest

from wind_up.constants import REANALYSIS_WD_COL
from wind_up.models import WindUpConfig
from wind_up.northing import apply_northing_corrections, calc_max_abs_north_errs
from wind_up.scada_funcs import scada_multi_index


def test_apply_northing_corrections(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config
    test_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_test_df.parquet")
    test_df.columns = test_df.columns.str.replace("test_", "")
    test_df["TurbineName"] = "LSA_T07"
    wf_df = test_df.copy()
    for wtg_name in ["LSA_T09", "LSA_T12", "LSA_T13", "LSA_T14"]:
        wtg_df = test_df.copy()
        wtg_df["TurbineName"] = wtg_name
        wf_df = pd.concat([wf_df, wtg_df])
    wf_df = scada_multi_index(wf_df)
    wf_df_after_northing = apply_northing_corrections(wf_df, cfg=cfg, north_ref_wd_col=REANALYSIS_WD_COL, plot_cfg=None)

    median_yaw_before_northing = wf_df.groupby("TurbineName")["YawAngleMean"].median()
    median_yaw_after_northing = wf_df_after_northing.groupby("TurbineName")["YawAngleMean"].median()
    assert median_yaw_before_northing.min() == pytest.approx(232.67355346679688)
    assert median_yaw_before_northing.max() == pytest.approx(232.67355346679688)
    assert median_yaw_after_northing["LSA_T07"] == pytest.approx(219.69126892089844)
    assert median_yaw_after_northing["LSA_T09"] == pytest.approx(245.49956512451172)
    assert median_yaw_after_northing["LSA_T12"] == pytest.approx(177.40547943115234)
    assert median_yaw_after_northing["LSA_T13"] == pytest.approx(235.22855377197266)
    assert median_yaw_after_northing["LSA_T14"] == pytest.approx(224.92881774902344)

    abs_north_errs_before_northing = calc_max_abs_north_errs(
        wf_df, north_ref_wd_col=REANALYSIS_WD_COL, timebase_s=cfg.timebase_s
    )
    abs_north_errs_after_northing = calc_max_abs_north_errs(
        wf_df_after_northing, north_ref_wd_col=REANALYSIS_WD_COL, timebase_s=cfg.timebase_s
    )
    assert abs_north_errs_before_northing.min() == pytest.approx(7.88920288085938)
    assert abs_north_errs_before_northing.max() == pytest.approx(7.88920288085938)
    assert abs_north_errs_after_northing["LSA_T07"] == pytest.approx(18.400006103515604)
    assert abs_north_errs_after_northing["LSA_T09"] == pytest.approx(23.889203)
    assert abs_north_errs_after_northing["LSA_T12"] == pytest.approx(162.918431)
    assert abs_north_errs_after_northing["LSA_T13"] == pytest.approx(10.889203)
    assert abs_north_errs_after_northing["LSA_T14"] == pytest.approx(12.400006)



from pathlib import Path

PROJECTROOT_DIR = Path(__file__).parents[1]
CONFIG_DIR = Path(__file__).parents[1] / "config"
TURBINE_DATA_DIR = Path(__file__).parents[1] / "input_data/turbine_data"
REANALYSIS_DIR = Path(__file__).parents[1] / "input_data/reanalysis"
TOGGLE_DIR = Path(__file__).parents[1] / "input_data/toggle"
OUTPUT_DIR = Path(__file__).parents[1] / "output"

RANDOM_SEED = 0
SCATTER_S = 1
SCATTER_ALPHA = 0.2
SCATTER_MARKERSCALE = 4


class DataColumns:
    turbine_name = "TurbineName"
    active_power_mean = "ActivePowerMean"
    active_power_sd = "ActivePowerSD"
    wind_speed_mean = "WindSpeedMean"
    wind_speed_sd = "WindSpeedSD"
    yaw_angle_mean = "YawAngleMean"
    yaw_angle_min = "YawAngleMin"
    yaw_angle_max = "YawAngleMax"
    pitch_angle_mean = "PitchAngleMean"
    gen_rpm_mean = "GenRpmMean"
    ambient_temp = "AmbientTemp"
    shutdown_duration = "ShutdownDuration"

    @classmethod
    def all(cls: type["DataColumns"]) -> list[str]:
        return [v for k, v in vars(cls).items() if not k.startswith("_") and k != "all"]


HOURS_PER_YEAR = 8766
DEFAULT_AIR_DENSITY = 1.22

TIMESTAMP_COL = "TimeStamp_StartFormat"
RAW_WINDSPEED_COL = "raw_WindSpeedMean"
RAW_POWER_COL = "raw_ActivePowerMean"
RAW_DOWNTIME_S_COL = "raw_ShutdownDuration"
RAW_YAWDIR_COL = "raw_YawAngleMean"

REANALYSIS_WS_COL = "reanalysis_ws"
REANALYSIS_WD_COL = "reanalysis_wd"
WINDFARM_YAWDIR_COL = "wf_yawdir"



import logging

import numpy as np
import pandas as pd
from geographiclib.geodesic import Geodesic
from tabulate import tabulate

from wind_up.constants import (
    DEFAULT_AIR_DENSITY,
    RAW_DOWNTIME_S_COL,
    RAW_POWER_COL,
    RAW_WINDSPEED_COL,
    TIMESTAMP_COL,
)
from wind_up.math_funcs import circ_diff
from wind_up.models import PlotConfig, TurbineType, WindUpConfig
from wind_up.plots.waking_state_plots import plot_waking_state_one_ttype_or_wtg
from wind_up.wind_funcs import calc_cp

logger = logging.getLogger(__name__)


def add_waking_state_one_ttype(
    wf_df: pd.DataFrame, *, ttype: TurbineType, timebase_s: int, plot_cfg: PlotConfig | None
) -> pd.DataFrame:
    wf_df = wf_df.copy()
    rated_power = ttype.rated_power_kw
    wf_df["waking"] = ~wf_df["ActivePowerMean"].isna()
    margin_for_waking_rated_power = 0.8
    wf_df["waking"] = wf_df["waking"] | (wf_df[RAW_POWER_COL] > margin_for_waking_rated_power * rated_power)
    margin_for_cp_calc = 0.3
    df_for_median_cp = wf_df[
        wf_df["waking"]
        & (wf_df[RAW_POWER_COL] > margin_for_cp_calc * rated_power)
        & (wf_df[RAW_POWER_COL] < (1 - margin_for_cp_calc) * rated_power)
    ]
    median_waking_cp = calc_cp(  # type: ignore[union-attr]
        df_for_median_cp["ActivePowerMean"],
        df_for_median_cp["WindSpeedMean"].clip(lower=1),
        DEFAULT_AIR_DENSITY,
        ttype.rotor_diameter_m,
    ).median()
    factor_for_waking_cp = 0.7
    wf_df["waking"] = wf_df["waking"] | (
        calc_cp(
            wf_df[RAW_POWER_COL],
            wf_df[RAW_WINDSPEED_COL].clip(lower=1),
            DEFAULT_AIR_DENSITY,
            ttype.rotor_diameter_m,
        )
        > factor_for_waking_cp * median_waking_cp
    )

    margin_for_not_waking_rated_power = 0.01
    wf_df["not_waking"] = wf_df[RAW_POWER_COL] < margin_for_not_waking_rated_power * rated_power
    wf_df["not_waking"] = wf_df["not_waking"] | (wf_df[RAW_DOWNTIME_S_COL] > timebase_s * factor_for_waking_cp)

    wf_df["not_waking"] = wf_df["not_waking"] & (~wf_df["waking"])
    wf_df["unknown_waking"] = (~wf_df["waking"]) & (~wf_df["not_waking"])

    if plot_cfg is not None:
        waking_frc = wf_df["waking"].sum() / len(wf_df)
        logger.info(f"{ttype.turbine_type} {waking_frc * 100:.1f}% of rows are waking")
        not_waking_frc = wf_df["not_waking"].sum() / len(wf_df)
        logger.info(
            f"{ttype.turbine_type} {not_waking_frc * 100:.1f}% of rows are not waking",
        )
        unknown_waking_frc = wf_df["unknown_waking"].sum() / len(wf_df)
        logger.info(
            f"{ttype.turbine_type} {unknown_waking_frc * 100:.1f}% of rows have unknown or partial waking",
        )
        plot_waking_state_one_ttype_or_wtg(wf_df=wf_df, ttype_or_wtg=ttype.turbine_type, plot_cfg=plot_cfg)
        if not plot_cfg.skip_per_turbine_plots:
            for wtg_name in wf_df.index.unique(level="TurbineName"):
                df_wtg = wf_df.loc[wtg_name]
                plot_waking_state_one_ttype_or_wtg(wf_df=df_wtg, ttype_or_wtg=wtg_name, plot_cfg=plot_cfg)
    return wf_df


def add_waking_state(cfg: WindUpConfig, wf_df: pd.DataFrame, plot_cfg: PlotConfig | None) -> pd.DataFrame:
    df_input = wf_df.copy()
    wf_df = pd.DataFrame()
    for ttype in cfg.list_unique_turbine_types():
        wtgs = cfg.list_turbine_ids_of_type(ttype)
        df_ttype = df_input.loc[wtgs]
        wf_df_ = add_waking_state_one_ttype(
            wf_df=df_ttype,
            ttype=ttype,
            timebase_s=cfg.timebase_s,
            plot_cfg=plot_cfg,
        )
        wf_df = pd.concat([wf_df, wf_df_])
    return wf_df.sort_index()


def calc_bearing(*, lat1: float, long1: float, lat2: float, long2: float) -> float:
    bearing_deg = Geodesic.WGS84.Inverse(lat1, long1, lat2, long2)["azi1"]
    return bearing_deg % 360


def calc_distance(*, lat1: float, long1: float, lat2: float, long2: float) -> float:
    return Geodesic.WGS84.Inverse(lat1, long1, lat2, long2)["s12"]


distance_and_bearing_cache: dict[tuple[float, float, float, float], tuple[float, float]] = {}


def get_distance_and_bearing(*, lat1: float, long1: float, lat2: float, long2: float) -> tuple[float, float]:
    # see if distance and bearing are in known_answers
    # if so, return them
    if (lat1, long1, lat2, long2) in distance_and_bearing_cache:
        distance_m, bearing_deg = distance_and_bearing_cache[(lat1, long1, lat2, long2)]
    else:
        distance_m = calc_distance(lat1=lat1, long1=long1, lat2=lat2, long2=long2)
        bearing_deg = calc_bearing(lat1=lat1, long1=long1, lat2=lat2, long2=long2)
        distance_and_bearing_cache[(lat1, long1, lat2, long2)] = (distance_m, bearing_deg)
    return distance_m, bearing_deg


def calc_iec_upwind_turbines(*, lat: float, long: float, wind_direction: float, cfg: WindUpConfig) -> list[str]:
    # find all turbines in cfg which are upwind of lat, long
    wind_direction = wind_direction % 360
    wtg_names = []
    bearings = []
    distances = []
    wtg_d_norms = []
    for wtg in cfg.asset.wtgs:
        wtg_lat = wtg.latitude
        wtg_long = wtg.longitude
        wtg_names.append(wtg.name)
        wtg_distance, wtg_bearing = get_distance_and_bearing(lat1=lat, long1=long, lat2=wtg_lat, long2=wtg_long)
        bearings.append(wtg_bearing)
        distances.append(wtg_distance)
        wtg_d_norms.append(wtg_distance / wtg.turbine_type.rotor_diameter_m)
    upwind_df = pd.DataFrame(
        data={
            "wtg_name": wtg_names,
            "bearing": bearings,
            "distance_m": distances,
            "distance_diameters": wtg_d_norms,
        },
    )
    upwind_df["disturbed_sector"] = 180.0
    ge_2_diameters = upwind_df["distance_diameters"] >= 2  # noqa PLR2004
    upwind_df.loc[ge_2_diameters, "disturbed_sector"] = (
        1.3 * np.rad2deg(np.arctan(2.5 / upwind_df.loc[ge_2_diameters, "distance_diameters"] + 0.15)) + 10
    )
    gt_20_diameters = upwind_df["distance_diameters"] > 20  # noqa PLR2004
    upwind_df.loc[gt_20_diameters, "disturbed_sector"] = 0

    upwind_df["relative_bearing"] = circ_diff(upwind_df["bearing"], wind_direction)
    upwind_df["iec_upwind"] = abs(upwind_df["relative_bearing"]) < upwind_df["disturbed_sector"] / 2

    upwind_turbine_list = list(upwind_df.query("iec_upwind")["wtg_name"].sort_values().values)
    if wind_direction % 90 == 0:
        logger.info(
            f"calc_iec_upwind_turbines lat={lat:.2f} long={long:.2f} "
            f"wind_dir={wind_direction:.0f} {upwind_turbine_list}",
        )
    return upwind_turbine_list


upwind_wtgs_cache: dict[tuple[float, float, float, str | None], list[str]] = {}


def lat_long_is_valid(lat: float, long: float) -> bool:
    return (-90 <= lat <= 90) and (-180 <= long <= 180)  # noqa PLR2004


def get_iec_upwind_turbines_one_latlong(
    *,
    lat: float,
    long: float,
    wind_direction: float,
    cfg: WindUpConfig,
    object_name: str | None = None,
) -> list[str]:
    if not lat_long_is_valid(lat, long):
        msg = f"lat={lat} long={long} is not a valid lat long"
        raise ValueError(msg)

    if (lat, long, wind_direction, object_name) in upwind_wtgs_cache:
        upwind_wtgs = upwind_wtgs_cache[(lat, long, wind_direction, object_name)]
    else:
        upwind_wtgs = calc_iec_upwind_turbines(lat=lat, long=long, wind_direction=wind_direction, cfg=cfg)
        if object_name is not None:
            upwind_wtgs = [x for x in upwind_wtgs if x.lower() != object_name.lower()]
        upwind_wtgs_cache[(lat, long, wind_direction, object_name)] = upwind_wtgs
    return upwind_wtgs


def get_iec_upwind_turbines(
    *,
    latlongs: list[tuple[float, float]],
    wind_direction: float,
    cfg: WindUpConfig,
    object_name: str | None = None,
) -> list[str]:
    upwind_wtgs = []
    for lat, long in latlongs:
        upwind_wtgs += get_iec_upwind_turbines_one_latlong(
            lat=lat,
            long=long,
            wind_direction=wind_direction,
            cfg=cfg,
            object_name=object_name,
        )
    return sorted(set(upwind_wtgs))


def calc_scen_name_from_wtgs_not_waking(wtgs_not_waking: list[str]) -> str:
    if len(wtgs_not_waking) < 1:
        msg = "wtgs_not_waking must have at least one element"
        raise ValueError(msg)
    return " ".join(wtgs_not_waking) + " offline"


def list_wtgs_offline_in_scen(waking_scenario: str) -> list[str]:
    return [x for x in waking_scenario.split(" ") if x != "offline"]


def add_waking_scen(
    *,
    test_name: str,
    ref_name: str,
    test_ref_df: pd.DataFrame,
    cfg: WindUpConfig,
    wf_df: pd.DataFrame,
    ref_wd_col: str,
    ref_lat: float,
    ref_long: float,
) -> pd.DataFrame:
    test_ref_df = test_ref_df.copy()
    test_ref_df["waking_scenario"] = "not calculated"
    try:
        test_wtg = next(x for x in cfg.asset.wtgs if x.name == test_name)
    except StopIteration as exc:
        msg = f"{test_name} not found in cfg.asset.wtgs"
        raise ValueError(msg) from exc

    all_turbines_waking = wf_df.groupby(TIMESTAMP_COL)["waking"].all().to_frame(name="all_turbines_waking")
    test_ref_df = test_ref_df.merge(all_turbines_waking, how="left", left_index=True, right_index=True)
    test_ref_df.loc[test_ref_df["all_turbines_waking"], "waking_scenario"] = "none offline"

    test_ref_df["rounded_wd"] = test_ref_df[ref_wd_col].round(0).mod(360)
    waking_df = wf_df[["waking", "not_waking", "unknown_waking"]].pivot_table(
        index=TIMESTAMP_COL,
        columns="TurbineName",
        aggfunc="max",
        observed=False,
    )

    for wd in sorted(test_ref_df["rounded_wd"].unique()):
        if wd >= 0:
            test_upwind_wtgs = get_iec_upwind_turbines(
                latlongs=test_wtg.get_latlongs(),
                wind_direction=wd,
                cfg=cfg,
                object_name=test_name,
            )
            if lat_long_is_valid(ref_lat, ref_long):
                ref_upwind_wtgs = get_iec_upwind_turbines(
                    latlongs=[(ref_lat, ref_long)],
                    wind_direction=wd,
                    cfg=cfg,
                    object_name=ref_name,
                )
            else:
                ref_upwind_wtgs = []
            upwind_wtgs = sorted(set(test_upwind_wtgs + ref_upwind_wtgs))
            if len(upwind_wtgs) == 0:
                test_ref_df.loc[test_ref_df["rounded_wd"] == wd, "waking_scenario"] = "none offline"
            else:
                for idx in test_ref_df[
                    (test_ref_df["rounded_wd"] == wd) & (test_ref_df["waking_scenario"] == "not calculated")
                ].index:
                    if waking_df.loc[idx, "unknown_waking"].loc[upwind_wtgs].any():
                        test_ref_df.loc[idx, "waking_scenario"] = "unknown"
                    elif waking_df.loc[idx, "waking"].loc[upwind_wtgs].all():
                        test_ref_df.loc[idx, "waking_scenario"] = "none offline"
                    else:
                        not_waking_series = waking_df.loc[idx, "not_waking"].loc[upwind_wtgs]
                        test_ref_df.loc[idx, "waking_scenario"] = calc_scen_name_from_wtgs_not_waking(
                            list(not_waking_series[not_waking_series].index),
                        )

    if (test_ref_df.dropna(subset=ref_wd_col)["waking_scenario"] == "not calculated").any():
        msg = "some rows with defined ref_wd_col have waking scenario = 'not calculated'"
        raise RuntimeError(msg)

    top_scens = test_ref_df.dropna(subset=ref_wd_col)["waking_scenario"].value_counts()[0:5]
    logger.info(f"top {len(top_scens)} {test_name} {ref_name} waking scenarios [%]:")
    _table = tabulate(
        (top_scens / len(test_ref_df.dropna(subset=ref_wd_col)) * 100).to_frame(),
        tablefmt="outline",
        floatfmt=".1f",
    )
    logger.info(f"{_table}")

    return test_ref_df.drop(columns=["rounded_wd", "all_turbines_waking"])



import pandas as pd
import pytest

from wind_up.conversions import ensure_utc

SAMPLE_NAIVE = pd.date_range(pd.Timestamp("2000-01-01T00:00"), periods=2)
SAMPLE_UTC = pd.date_range(pd.Timestamp("2000-01-01T00:00+00:00"), periods=2)
SAMPLE_OTHER = pd.date_range(pd.Timestamp("2000-01-01T05:00+05:00"), periods=2)


@pytest.mark.parametrize(
    ("t", "exp"),
    [
        (pd.DatetimeIndex(SAMPLE_NAIVE), pd.DatetimeIndex(SAMPLE_UTC)),
        (pd.DatetimeIndex(SAMPLE_UTC), pd.DatetimeIndex(SAMPLE_UTC)),
        (pd.DatetimeIndex(SAMPLE_OTHER), pd.DatetimeIndex(SAMPLE_UTC)),
        (pd.Timestamp("2000-01-01T00:00"), pd.Timestamp("2000-01-01T00:00+00:00")),
        (pd.Timestamp("2000-01-01T00:00+00:00"), pd.Timestamp("2000-01-01T00:00+00:00")),
        (pd.Timestamp("2000-01-01T05:00+05:00"), pd.Timestamp("2000-01-01T00:00+00:00")),
    ],
)
def test_ensure_utc(t: pd.Timestamp, exp: pd.Timestamp) -> None:
    actual = ensure_utc(t)

    if isinstance(t, pd.Timestamp):
        assert actual == exp
    else:
        pd.testing.assert_index_equal(actual, exp)



import matplotlib.pyplot as plt
import pandas as pd

from wind_up.constants import RAW_POWER_COL, RAW_WINDSPEED_COL, SCATTER_ALPHA, SCATTER_MARKERSCALE, SCATTER_S
from wind_up.models import PlotConfig, WindUpConfig


def plot_pc_one_ttype(pc: pd.DataFrame, ttype: str, plot_cfg: PlotConfig) -> None:
    plt.plot(pc["x_mean"], pc["y_mean"], marker=".", label="raw")
    plt.plot(pc["WindSpeedMean"], pc["pw_clipped"], marker=".", label="cooked", linestyle=":")
    plot_title = f"{ttype} mean power curve"
    plt.title(plot_title)
    plt.xlabel("WindSpeedMean [m/s]")
    plt.ylabel("pw_clipped [kW]")
    plt.legend()
    plt.grid()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        tdir = plot_cfg.plots_dir / ttype
        tdir.mkdir(exist_ok=True, parents=True)
        plt.savefig(plot_cfg.plots_dir / ttype / f"{plot_title}.png")
    plt.close()


def plot_pc_per_ttype(cfg: WindUpConfig, pc_per_ttype: dict, plot_cfg: PlotConfig) -> None:
    for ttype in cfg.list_unique_turbine_types():
        ttype_str = ttype.turbine_type
        plot_pc_one_ttype(pc=pc_per_ttype[ttype_str], ttype=ttype_str, plot_cfg=plot_cfg)


def plot_removed_data_one_ttype_or_wtg(
    t_df: pd.DataFrame,
    pc_df: pd.DataFrame,
    ttype_or_wtg: str,
    plot_cfg: PlotConfig,
) -> None:
    if RAW_WINDSPEED_COL in t_df.columns and RAW_POWER_COL in t_df.columns:
        plot_df = t_df[t_df["ActivePowerMean"].isna()]
        plt.figure()
        plt.scatter(
            plot_df[RAW_WINDSPEED_COL],
            plot_df[RAW_POWER_COL],
            s=SCATTER_S,
            alpha=SCATTER_ALPHA,
            label="removed data",
        )
        plt.plot(pc_df["WindSpeedMean"], pc_df["pw_clipped"], label="SCADA power curve", linestyle=":", color="C1")
        plot_title = f"{ttype_or_wtg} REMOVED DATA"
        plt.title(plot_title)
        plt.xlabel(f"{RAW_WINDSPEED_COL} [m/s]")
        plt.ylabel(f"{RAW_POWER_COL} [kW]")
        plt.legend(loc="best", markerscale=SCATTER_MARKERSCALE)
        plt.grid()
        plt.tight_layout()
        if plot_cfg.show_plots:
            plt.show()
        if plot_cfg.save_plots:
            plt.savefig(plot_cfg.plots_dir / ttype_or_wtg / f"{plot_title}.png")
        plt.close()
    else:
        msg = f"cannot plot removed data without columns {RAW_WINDSPEED_COL} and {RAW_POWER_COL}"
        raise ValueError(msg)


def plot_removed_data_per_ttype_and_wtg(
    cfg: WindUpConfig,
    wf_df: pd.DataFrame,
    pc_per_ttype: dict,
    plot_cfg: PlotConfig,
) -> None:
    for ttype in cfg.list_unique_turbine_types():
        ttype_str = ttype.turbine_type
        wtg_names = cfg.list_turbine_ids_of_type(ttype)
        df_ttype = wf_df.loc[wtg_names]
        plot_removed_data_one_ttype_or_wtg(
            t_df=df_ttype,
            pc_df=pc_per_ttype[ttype_str],
            ttype_or_wtg=ttype_str,
            plot_cfg=plot_cfg,
        )
        if not plot_cfg.skip_per_turbine_plots:
            for wtg_name in wtg_names:
                plot_removed_data_one_ttype_or_wtg(
                    t_df=df_ttype.loc[[wtg_name]],
                    pc_df=pc_per_ttype[ttype_str],
                    ttype_or_wtg=wtg_name,
                    plot_cfg=plot_cfg,
                )



import datetime as dt
import logging

import pandas as pd

from wind_up.constants import PROJECTROOT_DIR, TIMESTAMP_COL, TURBINE_DATA_DIR
from wind_up.models import WindUpConfig
from wind_up.result_manager import result_manager

logger = logging.getLogger(__name__)


def calc_last_xmin_datetime_in_month(d: dt.datetime, td: pd.Timedelta) -> pd.Timestamp:
    return pd.Timestamp(
        pd.offsets.MonthBegin(normalize=True).rollforward(d + td) - td,
    )


def add_smart_lat_long_to_cfg(md: pd.DataFrame, cfg: WindUpConfig) -> WindUpConfig:
    # confirm md has the same turbines as cfg.asset.wtgs
    if set(md["Name"].values) != {x.name for x in cfg.asset.wtgs}:
        msg = "md has different turbines than cfg.asset.wtgs"
        raise ValueError(msg)
    # add md lat longs to cfg
    for wtg in cfg.asset.wtgs:
        wtg.latitude = float(md.loc[md["Name"] == wtg.name, "Latitude"].iloc[0])
        wtg.longitude = float(md.loc[md["Name"] == wtg.name, "Longitude"].iloc[0])
    return cfg


def load_smart_md_from_file(asset_name: str, *, test_mode: bool = False) -> pd.DataFrame:
    if test_mode:
        md_dir = PROJECTROOT_DIR / "tests" / "test_data" / "smart_data" / asset_name
    else:
        md_dir = TURBINE_DATA_DIR / asset_name
    return pd.read_csv(md_dir / f"{asset_name}_md.csv", index_col=0)


def calc_month_list_and_time_info(
    asset_name: str,
    first_datetime_utc_start: pd.Timestamp,
    last_datetime_utc_start: pd.Timestamp,
    md: pd.DataFrame,
    timebase_s: int,
) -> tuple[list[dt.datetime], dt.datetime, str, str]:
    if len(md["TimeSpanMinutes"].unique()) > 1 or (60 * md["TimeSpanMinutes"].unique()[0]) != timebase_s:
        msg = (
            f"TimeSpanMinutes not consistent with timebase {timebase_s=} for all SMART metadata in "
            f"{asset_name}: {md['TimeSpanMinutes'].unique()=}"
        )
        result_manager.warning(msg)

    # make sure there is only one TimeZone and TimeFormat
    if len(md["TimeZone"].unique()) > 1:
        msg = f"more than one TimeZone found in SMART metadata for {asset_name}: {md['TimeZone'].unique()}"
        raise ValueError(msg)
    if len(md["TimeFormat"].unique()) > 1:
        msg = f"more than one TimeFormat found in SMART metadata for {asset_name}: {md['TimeFormat'].unique()}"
        raise ValueError(
            msg,
        )

    smart_tz = md.loc[:, "TimeZone"].iloc[0]
    smart_tf = md.loc[:, "TimeFormat"].iloc[0]
    if smart_tf not in ("Start", "End"):
        msg = f"TimeFormat not Start or End in SMART metadata for {asset_name}: {smart_tf}"
        raise ValueError(msg)

    first_smart_dt_no_tz = first_datetime_utc_start.tz_convert(smart_tz).tz_localize(None).to_pydatetime()
    last_smart_dt_no_tz = last_datetime_utc_start.tz_convert(smart_tz).tz_localize(None).to_pydatetime()
    if smart_tf == "End":
        first_smart_dt_no_tz = first_smart_dt_no_tz + pd.Timedelta(seconds=timebase_s)
        last_smart_dt_no_tz = last_smart_dt_no_tz + pd.Timedelta(seconds=timebase_s)

    month_start_list_no_tz = pd.date_range(
        first_smart_dt_no_tz.strftime("%Y-%m-%d"),
        last_smart_dt_no_tz.strftime("%Y-%m-%d"),
        freq="MS",
    ).to_list()
    if len(month_start_list_no_tz) == 0:
        month_start_list_no_tz.insert(0, pd.Timestamp(first_smart_dt_no_tz.strftime("%Y-%m-%d")))
    if first_smart_dt_no_tz < month_start_list_no_tz[0]:
        month_start_list_no_tz.insert(0, pd.Timestamp(first_smart_dt_no_tz.strftime("%Y-%m-%d")))
    return month_start_list_no_tz, last_smart_dt_no_tz, smart_tz, smart_tf


def check_and_convert_scada_raw(
    scada_raw: pd.DataFrame,
    *,
    scada_data_timezone: str,
    scada_data_time_format: str,
    first_datetime_utc_start: pd.Timestamp,
    last_datetime_utc_start: pd.Timestamp,
    timebase_s: int,
) -> pd.DataFrame:
    if scada_raw["TurbineName"].isna().any():
        msg = f"NaNs in TurbineName column of scada_raw: {scada_raw['TurbineName'].isna().sum()}"
        raise RuntimeError(msg)
    scada_raw["TurbineName"] = scada_raw["TurbineName"].astype("category")

    scada_raw = scada_raw.sort_index()

    # save a copy of scada_raw index as a column
    scada_raw["smart_dtTimeStamp"] = scada_raw.index

    # convert scada_raw index from SMART tz to the wind-up timezone (UTC) and Start format
    if not isinstance(scada_raw.index, pd.DatetimeIndex):
        msg = f"scada_raw.index is not a pd.DatetimeIndex: {type(scada_raw.index)}"
        raise TypeError(msg)

    if scada_raw.index.tzinfo is not None:
        scada_raw.index = scada_raw.index.tz_localize(None)
    scada_raw.index = scada_raw.index.tz_localize(scada_data_timezone).tz_convert("UTC")
    if scada_data_time_format == "End":
        scada_raw.index = scada_raw.index - pd.Timedelta(seconds=timebase_s)
    scada_raw.index = scada_raw.index.rename(TIMESTAMP_COL)

    # clip to originally requested datetime range
    scada_raw = scada_raw.loc[first_datetime_utc_start:last_datetime_utc_start]

    # verify there is only one row per turbine per timestamp
    scada_raw.set_index([scada_raw.index, "TurbineName"], verify_integrity=True)

    turbine_rows = scada_raw.groupby("TurbineName", observed=False)["TurbineName"].count().to_frame()
    rows_per_turbine = turbine_rows.max().iloc[0]
    if rows_per_turbine != turbine_rows.min().iloc[0]:
        msg = f"turbines have different number of rows: {rows_per_turbine} != {turbine_rows.min().iloc[0]}"
        result_manager.warning(msg)
        logger.info("attempting to repair")
        rows_before = len(scada_raw)
        scada_raw_reset = scada_raw.reset_index()
        unique_wtgs = scada_raw_reset["TurbineName"].unique()
        complete_index = pd.date_range(
            start=scada_raw.index.min(),
            end=scada_raw.index.max(),
            freq=pd.Timedelta(seconds=timebase_s),
        )
        reindexed_df = pd.DataFrame()
        for wtg in unique_wtgs:
            temp_df = scada_raw_reset[scada_raw_reset["TurbineName"] == wtg]
            temp_df = temp_df.set_index(TIMESTAMP_COL)
            temp_df = temp_df.reindex(complete_index)
            temp_df["TurbineName"] = wtg
            reindexed_df = pd.concat([reindexed_df, temp_df])
        reindexed_df.index.name = TIMESTAMP_COL
        rows_after = len(reindexed_df)

        turbine_rows = reindexed_df.groupby("TurbineName", observed=False)["TurbineName"].count().to_frame()
        new_rows_per_turbine = turbine_rows.max().iloc[0]
        if new_rows_per_turbine == turbine_rows.min().iloc[0] and new_rows_per_turbine == rows_per_turbine:
            logger.info(f"repair successful. rows before: {rows_before}, rows after: {rows_after}")
            scada_raw = reindexed_df
        else:
            msg = f"turbines have different number of rows: {new_rows_per_turbine} != {turbine_rows.min().iloc[0]}"
            raise RuntimeError(msg)
    rows_per_hour = 3600 / timebase_s
    logger.info(
        f"loaded {len(turbine_rows)} turbines, {rows_per_turbine / rows_per_hour / 24 / 365.25:.1f} years per turbine"
    )
    return scada_raw


def load_smart_scada_and_md_from_file(
    asset_name: str,
    scada_df: pd.DataFrame,
    metadata_df: pd.DataFrame,
    *,
    first_datetime_utc_start: pd.Timestamp,
    last_datetime_utc_start: pd.Timestamp,
    timebase_s: int,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    logger.info(
        f"running load_smart_scada_and_md_from_file for {first_datetime_utc_start} to {last_datetime_utc_start}"
    )
    md = metadata_df

    month_start_list_no_tz, last_smart_dt_no_tz, smart_tz, smart_tf = calc_month_list_and_time_info(
        asset_name=asset_name,
        first_datetime_utc_start=first_datetime_utc_start,
        last_datetime_utc_start=last_datetime_utc_start,
        md=md,
        timebase_s=timebase_s,
    )
    scada_raw = check_and_convert_scada_raw(
        scada_df,
        scada_data_timezone=smart_tz,
        scada_data_time_format=smart_tf,
        first_datetime_utc_start=first_datetime_utc_start,
        last_datetime_utc_start=last_datetime_utc_start,
        timebase_s=timebase_s,
    )

    logger.info(f"finished load_smart_scada_and_md for {first_datetime_utc_start} to {last_datetime_utc_start}")
    return scada_raw, md



import datetime as dt
import logging
from pathlib import Path

import numpy as np
import pandas as pd
import yaml
from pydantic import BaseModel, Field, model_validator

from wind_up.constants import OUTPUT_DIR
from wind_up.yaml_loader import Loader, construct_include

logger = logging.getLogger(__name__)


class PlotConfig(BaseModel):
    show_plots: bool = Field(default=False, description="Show plots in interactive window")
    save_plots: bool = Field(default=True, description="Save plots to file")
    skip_per_turbine_plots: bool = Field(default=False, description="If True skip per turbine plots")
    plots_dir: Path = Field(description="Directory to save plots to")

    @model_validator(mode="after")
    def make_plots_dir(self: "PlotConfig") -> "PlotConfig":
        self.plots_dir.mkdir(parents=True, exist_ok=True)
        return self


class TurbineType(BaseModel):
    turbine_type: str = Field(
        description="Turbine type description",
        min_length=2,
    )
    rotor_diameter_m: float = Field(description="Rotor diameter in meters", gt=0, examples=[100])
    rated_power_kw: float = Field(description="Rated power in kW", gt=0, examples=[3000])
    cutout_ws_mps: float = Field(default=25, description="Cutout wind speed in m/s", gt=0, examples=[25])
    normal_operation_pitch_range: tuple[float, float] = Field(
        description="Normal operation pitch range in degrees",
        examples=[[-10, 40]],
    )
    normal_operation_genrpm_range: tuple[float, float] = Field(
        description="Normal operation generator rpm range",
        examples=[[800, 1600]],
    )
    rpm_v_pw_margin_factor: float = Field(
        default=0.05,
        description="Factor to multiply max RPM by to get rpm margin for rpm vs power filtering",
        examples=[0.05, 0.1],
    )
    pitch_to_stall: bool = Field(
        default=False,
        description="If False then pitch filter will eliminate data below the threshold rather than above",
    )


class Turbine(BaseModel):
    name: str = Field(description="Turbine name", min_length=2)
    turbine_type: TurbineType
    latitude: float = Field(default=np.nan, ge=-90, le=90)
    longitude: float = Field(default=np.nan, ge=-180, le=180)

    def get_latlongs(self: "Turbine") -> list[tuple[float, float]]:
        return [(self.latitude, self.longitude)]


class MastOrLidar(BaseModel):
    name: str = Field(description="Object name", min_length=2)
    latitude: float = Field(default=np.nan, ge=-90, le=90)
    longitude: float = Field(default=np.nan, ge=-180, le=180)
    data_file_name: str = Field(
        min_length=2,
        description="Name of data timeseries file",
    )
    wind_speed_column: str = Field(min_length=2, description="Name of wind speed column in data timeseries file")
    wind_direction_column: str = Field(
        min_length=2,
        description="Name of wind direction column in data timeseries file",
    )


class Asset(BaseModel):
    name: str = Field(description="Asset Name", min_length=2, examples=["Kelmarsh"])
    wtgs: list[Turbine]
    masts_and_lidars: list[MastOrLidar] = Field(description="list of mast and LiDAR objects, if any", default=[])


class Toggle(BaseModel):
    name: str = Field(description="Name of toggle signal to use in plots", min_length=2, default="toggle")
    toggle_file_per_turbine: bool = Field(
        description="Is there one toggle timeseries file per turbine, or one for the asset?",
    )
    toggle_filename: str = Field(
        min_length=2,
        description="Name of toggle timeseries file, can use {wtg} if toggle_file_per_turbine is True",
    )
    detrend_data_selection: str = Field(
        description="Method for selecting directional detrending data",
        default="do_not_use_toggle_test",
        examples=["do_not_use_toggle_test", "use_toggle_off_data"],
    )
    pairing_filter_method: str = Field(
        description="Method to use for pairing filter",
        examples=["none", "one_to_one", "any_within_timedelta"],
        default="none",
    )
    pairing_filter_timedelta_seconds: int = Field(
        description="Time delta in seconds to use for pairing filter",
        default=50 * 60,
    )
    toggle_change_settling_filter_seconds: int = Field(
        description="Time delta in seconds filter out after a toggle state change",
        default=10 * 60,
    )


class PrePost(BaseModel):
    pre_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in pre-upgrade analysis, UTC Start format",
    )
    pre_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in pre-upgrade analysis, UTC Start format",
    )
    post_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in post-upgrade analysis, UTC Start format",
    )
    post_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in post-upgrade analysis, UTC Start format",
    )


class WindUpConfig(BaseModel):
    assessment_name: str = Field(
        min_length=2,
        description="Name used for assessment output folder",
    )
    timebase_s: int = Field(
        default=10 * 60,
        description="Timebase in seconds for SCADA data, other data is converted to this timebase",
    )
    ignore_turbine_anemometer_data: bool = Field(
        default=False,
        description="If true do not use turbine anemometer data for anything",
    )
    require_test_wake_free: bool = Field(
        default=False,
        description="Remove data where the test turbine has any upwind turbines",
    )
    require_ref_wake_free: bool = Field(
        default=False,
        description="Remove data where the reference has any upwind turbines",
    )
    detrend_min_hours: int = Field(
        default=24,
        description="Minimum number of hours to use in directional detrending",
    )
    ref_wd_filter: list[float] | None = Field(
        default=None,
        description="Wind direction filter for reference data; only data within this range is used",
        examples=[[195, 241]],
    )
    ref_hod_filter: list[float] | None = Field(
        default=None,
        description="Hour of day filter for reference data; only data within this range is used",
        examples=[[18, 6]],
    )
    filter_all_test_wtgs_together: bool = Field(
        description="If True any row filtered for one test wtg is filtered for all test wtgs. "
        "Recommended for wake steering analysis.",
        default=False,
    )
    use_lt_distribution: bool = Field(
        description="If True the long term distribution is calculated and used for uplift calculation", default=True
    )
    out_dir: Path = Field(description="Directory to save assessment output to")
    test_wtgs: list[Turbine] = Field(description="List of test Turbine ids")
    ref_wtgs: list[Turbine] = Field(default=[], description="List of reference Turbine ids")
    ref_super_wtgs: list[str] = Field(
        default=[],
        description="List of reference super Turbine ids",
    )
    non_wtg_ref_names: list[str] = Field(
        default=[],
        description="List of non turbine references",
    )
    upgrade_first_dt_utc_start: dt.datetime = Field(
        description="First time when test turbine is upgraded, UTC Start format",
    )
    analysis_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in upgrade analysis, UTC Start format",
    )
    analysis_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in upgrade analysis, UTC Start format",
    )
    lt_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in long term analysis, UTC Start format",
    )
    lt_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in long term analysis, UTC Start format",
    )
    detrend_first_dt_utc_start: dt.datetime = Field(
        description="First time to use in detrend analysis, UTC Start format",
    )
    detrend_last_dt_utc_start: dt.datetime = Field(
        description="Last time to use in detrend analysis, UTC Start format",
    )
    years_offset_for_pre_period: int | None = Field(
        description="How many years to go back in time to begin pre upgrade dataset",
        ge=0,
        examples=[None, 1],
        default=None,
    )
    years_for_lt_distribution: int = Field(
        description="How many years to use in long term analysis",
        ge=0,
        examples=[3, 5],
    )
    years_for_detrend: int = Field(
        description="How many years to use in directional detrending",
        ge=0,
        examples=[1, 2],
    )
    ws_bin_width: float = Field(description="Wind speed bin width in m/s", gt=0, examples=[0.5, 1])
    bootstrap_runs_override: int | None = Field(
        description="Number of bootstrap runs to use, if None then calculated",
        ge=0,
        examples=[None, 1000],
        default=None,
    )
    reanalysis_method: str = Field(
        default="node_with_best_ws_corr",
        description="Method to use for reanalysis (e.g. ERA5) selection",
        examples=["node_with_best_ws_corr"],
    )
    smart_missing_data_fields: list[str] = Field(
        default=[],
        description="List of SCADA fields not available for this asset",
        examples=[["YawAngleMin", "YawAngleMax"]],
    )
    asset: Asset
    exclusion_periods_utc: list[tuple[str, dt.datetime, dt.datetime]] = Field(
        default=[],
        description="Turbine id or 'ALL', first UTC timestamp to exclude, last UTC timestamp to exclude",
        examples=[
            "ALL",
            dt.datetime(2021, 9, 21, 12, 40, tzinfo=dt.timezone.utc),
            dt.datetime(2021, 9, 29, 18, 10, tzinfo=dt.timezone.utc),
        ],
    )
    yaw_data_exclusions_utc: list[tuple[str, dt.datetime, dt.datetime]] = Field(
        default=[],
        description="Turbine id or 'ALL', first UTC start format timestamp to exclude yaw data,"
        " last UTC timestamp to exclude yaw data",
        examples=[
            "ALL",
            dt.datetime(2021, 9, 21, 12, 40, tzinfo=dt.timezone.utc),
            dt.datetime(2021, 9, 29, 18, 10, tzinfo=dt.timezone.utc),
        ],
    )
    optimize_northing_corrections: bool = Field(
        default=False,
        description="run northing correction optimization",
    )
    northing_corrections_utc: list[tuple[str, dt.datetime, float]] = Field(
        default=[],
        description="Turbine id, first UTC timestamp to correct, correction to add to yaw data",
        examples=["SMV1", dt.datetime(2021, 9, 21, 12, 40, tzinfo=dt.timezone.utc), 15],
    )
    toggle: Toggle | None = None
    prepost: PrePost | None = None

    @model_validator(mode="after")
    def check_years_offset_for_pre_period(self: "WindUpConfig") -> "WindUpConfig":
        if self.toggle is None and self.years_offset_for_pre_period is None:
            msg = "toggle is None and years_offset_for_pre_period is None"
            raise ValueError(msg)
        return self

    @model_validator(mode="after")
    def check_first_datetime_before_last(self: "WindUpConfig") -> "WindUpConfig":
        if self.upgrade_first_dt_utc_start >= self.analysis_last_dt_utc_start:
            msg = "upgrade_first_datetime must be before last_useable_datetime"
            raise ValueError(msg)
        return self

    @model_validator(mode="after")
    def check_non_wtg_ref_names(self: "WindUpConfig") -> "WindUpConfig":
        for non_wtg_ref in self.non_wtg_ref_names:
            if non_wtg_ref == "reanalysis":
                if len(self.reanalysis_method) < 1:
                    msg = "reanalysis is in non_wtg_ref_names but reanalysis_method is not valid"
                    raise ValueError(msg)
            elif len([x for x in self.asset.masts_and_lidars if x.name == non_wtg_ref]) != 1:
                msg = f"there is not exactly 1 non_wtg_ref {non_wtg_ref} in cfg.asset.masts_and_lidars"
                raise ValueError(msg)
        return self

    @model_validator(mode="after")
    def print_summary(self: "WindUpConfig") -> "WindUpConfig":
        logger.info(f"loaded WindUpConfig assessment_name: {self.assessment_name}")
        dt_fmt = "%Y-%m-%d %H:%M"
        if self.toggle is not None:
            dt_rng_start = self.analysis_first_dt_utc_start.strftime(dt_fmt)
            dt_rng_end = (self.analysis_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
            logger.info(
                f"toggle analysis period (UTC): {dt_rng_start} to {dt_rng_end}",
            )
        elif self.prepost is not None:
            dt_rng_start = self.prepost.pre_first_dt_utc_start.strftime(dt_fmt)
            dt_rng_end = (self.prepost.pre_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
            logger.info(
                f"pre analysis period (UTC): {dt_rng_start} to {dt_rng_end}",
            )
            dt_rng_start = self.prepost.post_first_dt_utc_start.strftime(dt_fmt)
            dt_rng_end = (self.prepost.post_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
            logger.info(
                f"post analysis period (UTC): {dt_rng_start} to {dt_rng_end}",
            )
        else:
            msg = "toggle and prepost are both set to None"
            raise RuntimeError(msg)
        dt_rng_start = self.lt_first_dt_utc_start.strftime(dt_fmt)
        dt_rng_end = (self.lt_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
        logger.info(f"long term period (UTC): {dt_rng_start} to {dt_rng_end}")
        dt_rng_start = self.detrend_first_dt_utc_start.strftime(dt_fmt)
        dt_rng_end = (self.detrend_last_dt_utc_start + pd.Timedelta(seconds=self.timebase_s)).strftime(dt_fmt)
        logger.info(f"detrend period (UTC): {dt_rng_start} to {dt_rng_end}")
        return self

    @classmethod
    def from_yaml(cls, file_path: Path) -> "WindUpConfig":  # noqa ANN102
        yaml.add_constructor("!include", construct_include, Loader)
        with Path.open(file_path) as f:
            cfg_dct = yaml.load(f, Loader)  # noqa S506

        cfg_dct["out_dir"] = OUTPUT_DIR / cfg_dct["assessment_name"]

        # ensure datetimes are UTC aware
        cfg_dct["upgrade_first_dt_utc_start"] = pd.Timestamp(cfg_dct["upgrade_first_dt_utc_start"], tz="UTC")
        cfg_dct["analysis_last_dt_utc_start"] = pd.Timestamp(cfg_dct["analysis_last_dt_utc_start"], tz="UTC")
        if cfg_dct.get("exclusion_periods_utc", None) is not None:
            for x in cfg_dct["exclusion_periods_utc"]:
                x[1] = pd.Timestamp(x[1], tz="UTC")
                x[2] = pd.Timestamp(x[2], tz="UTC")
        if cfg_dct.get("yaw_data_exclusions_utc", None) is not None:
            for x in cfg_dct["yaw_data_exclusions_utc"]:
                x[1] = pd.Timestamp(x[1], tz="UTC")
                x[2] = pd.Timestamp(x[2], tz="UTC")
        if cfg_dct.get("northing_corrections_utc", None) is not None:
            for x in cfg_dct["northing_corrections_utc"]:
                x[1] = pd.Timestamp(x[1], tz="UTC")

        # calculate analysis and long term datetimes
        test_is_toggle = cfg_dct.get("toggle", None) is not None
        if test_is_toggle:
            cfg_dct["analysis_first_dt_utc_start"] = cfg_dct["upgrade_first_dt_utc_start"]
        else:
            pre_post_dict = {
                "post_first_dt_utc_start": cfg_dct["upgrade_first_dt_utc_start"],
                "post_last_dt_utc_start": cfg_dct["analysis_last_dt_utc_start"],
                "pre_first_dt_utc_start": cfg_dct["upgrade_first_dt_utc_start"]
                - dt.timedelta(days=(365.25 * cfg_dct["years_offset_for_pre_period"])),
                "pre_last_dt_utc_start": cfg_dct["analysis_last_dt_utc_start"]
                - dt.timedelta(days=(365.25 * cfg_dct["years_offset_for_pre_period"])),
            }
            cfg_dct["prepost"] = PrePost.model_validate(pre_post_dict)
            cfg_dct["analysis_first_dt_utc_start"] = cfg_dct["prepost"].pre_first_dt_utc_start
        cfg_dct["lt_last_dt_utc_start"] = (
            cfg_dct["upgrade_first_dt_utc_start"] - dt.timedelta(days=7)  # go back 1 week for buffer before toggling
            if test_is_toggle
            else cfg_dct["prepost"].pre_last_dt_utc_start
        )
        cfg_dct["lt_first_dt_utc_start"] = cfg_dct["lt_last_dt_utc_start"] - dt.timedelta(
            days=(cfg_dct["years_for_lt_distribution"] * 365.25),
        )
        cfg_dct["detrend_last_dt_utc_start"] = cfg_dct["lt_last_dt_utc_start"]
        if test_is_toggle and cfg_dct["toggle"].get("detrend_data_selection", None) == "use_toggle_off_data":
            cfg_dct["detrend_last_dt_utc_start"] = cfg_dct["analysis_last_dt_utc_start"]
        cfg_dct["detrend_first_dt_utc_start"] = cfg_dct["detrend_last_dt_utc_start"] - dt.timedelta(
            days=(cfg_dct["years_for_detrend"] * 365.25),
        )

        # check each test_wtg exists in asset.wtgs
        for x in cfg_dct["test_wtgs"]:
            if x not in cfg_dct["asset"]["wtgs"]:
                msg = f"test_wtg {x} does not exist in asset.wtgs"
                raise ValueError(msg)

        # if ref_wtgs exist, check each ref_wtg exists in asset.wtgs
        if cfg_dct.get("ref_wtgs", None) is not None:
            for x in cfg_dct["ref_wtgs"]:
                if x not in cfg_dct["asset"]["wtgs"]:
                    msg = f"ref_wtg {x} does not exist in asset.wtgs"
                    raise ValueError(msg)

        # change lists of strings into list of turbines
        if len(cfg_dct["asset"]["turbine_types"]) == 1:
            cfg_dct["asset"]["turbine_types"] = cfg_dct["asset"]["turbine_types"] * len(
                cfg_dct["asset"]["wtgs"],
            )
        tt_list = [TurbineType.model_validate(tt) for tt in cfg_dct["asset"]["turbine_types"]]
        cfg_dct["asset"]["wtgs"] = [
            Turbine(name=x, turbine_type=tt) for x, tt in zip(cfg_dct["asset"]["wtgs"], tt_list, strict=True)
        ]
        test_wtg_list = []
        for x in cfg_dct["test_wtgs"]:
            for wtg in cfg_dct["asset"]["wtgs"]:
                if x == wtg.name:
                    test_wtg_list.append(wtg)
                    continue
        cfg_dct["test_wtgs"] = test_wtg_list
        if cfg_dct.get("ref_wtgs", None) is not None:
            ref_wtg_list = []
            for x in cfg_dct["ref_wtgs"]:
                for wtg in cfg_dct["asset"]["wtgs"]:
                    if x == wtg.name:
                        ref_wtg_list.append(wtg)
                        continue
            cfg_dct["ref_wtgs"] = ref_wtg_list
        elif cfg_dct.get("non_wtg_ref_names", None) is not None:
            if len(cfg_dct["non_wtg_ref_names"]) < 1:
                msg = "ref_wtgs or non_wtg_ref_names must have more than one element"
                raise ValueError(msg)
        else:
            msg = "ref_wtgs or non_wtg_ref_names must be specified"
            raise ValueError(msg)
        return WindUpConfig.model_validate(cfg_dct)

    def get_max_rated_power(self: "WindUpConfig") -> float:
        return max(x.turbine_type.rated_power_kw for x in self.asset.wtgs)

    def list_unique_turbine_types(self: "WindUpConfig") -> list["TurbineType"]:
        unique_names = sorted({x.turbine_type.turbine_type for x in self.asset.wtgs})
        unique_turbine_types = []
        for name in unique_names:
            unique_turbine_types.append(
                next(x.turbine_type for x in self.asset.wtgs if x.turbine_type.turbine_type == name),
            )
        return unique_turbine_types

    def list_turbine_ids_of_type(self: "WindUpConfig", ttype: "TurbineType") -> list[str]:
        return [x.name for x in self.asset.wtgs if x.turbine_type == ttype]

    def get_normal_operation_genrpm_range(self: "WindUpConfig", ttype: "TurbineType") -> tuple[float, float]:
        return next(x.turbine_type.normal_operation_genrpm_range for x in self.asset.wtgs if x.turbine_type == ttype)

    def get_normal_operation_pitch_range(self: "WindUpConfig", ttype: "TurbineType") -> tuple[float, float]:
        return next(x.turbine_type.normal_operation_pitch_range for x in self.asset.wtgs if x.turbine_type == ttype)



import logging
import math
from dataclasses import dataclass
from pathlib import Path

import pandas as pd

from wind_up.caching import with_pickle_cache
from wind_up.constants import REANALYSIS_WD_COL
from wind_up.models import PlotConfig, WindUpConfig
from wind_up.northing import add_wf_yawdir, apply_northing_corrections
from wind_up.optimize_northing import auto_northing_corrections
from wind_up.reanalysis_data import ReanalysisDataset, add_reanalysis_data
from wind_up.scada_funcs import filter_scada_df, get_raw_scada_and_cfg_from_file
from wind_up.scada_power_curve import calc_pc_and_rated_ws
from wind_up.smart_data import add_smart_lat_long_to_cfg
from wind_up.waking_state import add_waking_state
from wind_up.ws_est import add_ws_est

logger = logging.getLogger(__name__)


class PrePostSplitter:
    def __init__(self, cfg: WindUpConfig, toggle_df: pd.DataFrame | None = None):
        self.cfg = cfg
        self.toggle_df = toggle_df

    def split(self, df: pd.DataFrame, test_wtg_name: str) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        if (self.cfg.prepost is not None) and self.cfg.toggle is None:
            test_df = df.copy()
            pre_df = df[df.index <= self.cfg.prepost.pre_last_dt_utc_start].copy()
            post_df = df[df.index >= self.cfg.prepost.post_first_dt_utc_start].copy()
        elif (self.cfg.prepost is None) and self.cfg.toggle is not None:
            if not isinstance(self.toggle_df, pd.DataFrame):
                raise ValueError("toggle_df must be a pd.DataFrame")  # noqa TRY003
            test_df = add_toggle_signals(df, toggle_df=self.toggle_df, wtg_name=test_wtg_name, cfg=self.cfg)
            test_df = test_df.rename(columns={"toggle_off": "test_toggle_off", "toggle_on": "test_toggle_on"})
            pre_df = test_df[test_df["test_toggle_off"].fillna(value=False)].copy()
            post_df = test_df[test_df["test_toggle_on"].fillna(value=False)].copy()
        else:
            msg = "cfg must have exactly one of prepost or toggle"
            raise ValueError(msg)
        pre_df = pre_df[pre_df.index >= self.cfg.analysis_first_dt_utc_start]
        post_df = post_df[post_df.index >= self.cfg.analysis_first_dt_utc_start]
        return test_df, pre_df, post_df


def add_toggle_signals(
    input_df: pd.DataFrame, toggle_df: pd.DataFrame, wtg_name: str, cfg: WindUpConfig
) -> pd.DataFrame:
    toggle_df = toggle_df.copy()
    if cfg.toggle is None:
        msg = "add_toggle_signals cannot be run if cfg.toggle is None"
        raise ValueError(msg)
    if (toggle_df["toggle_on"] & toggle_df["toggle_off"]).any():
        msg = "toggle_on and toggle_off cannot be True at the same time"
        raise RuntimeError(msg)
    if toggle_df.index.nlevels == 2:  # noqa PLR2004
        toggle_df = toggle_df.loc[wtg_name]
    if cfg.toggle.toggle_change_settling_filter_seconds > 0:
        # apply toggle_change_settling_filter_seconds
        # first find toggle change times
        toggle_on_rows_before = toggle_df["toggle_on"].sum()
        toggle_off_rows_before = toggle_df["toggle_off"].sum()

        toggle_on_change_times = toggle_df.index[(toggle_df["toggle_on"] & toggle_df.shift(1)["toggle_off"])]
        toggle_off_change_times = toggle_df.index[(toggle_df["toggle_off"] & toggle_df.shift(1)["toggle_on"])]
        toggle_change_times = pd.DatetimeIndex(
            toggle_on_change_times.to_list() + toggle_off_change_times.to_list(),
        ).sort_values()
        rows_to_filter = math.ceil(cfg.toggle.toggle_change_settling_filter_seconds / cfg.timebase_s)
        while rows_to_filter > 0:
            toggle_df.loc[toggle_change_times, "toggle_on"] = False
            toggle_df.loc[toggle_change_times, "toggle_off"] = False
            toggle_change_times = toggle_change_times + pd.Timedelta(seconds=cfg.timebase_s)
            rows_to_filter -= 1

        toggle_on_rows_after = toggle_df["toggle_on"].sum()
        toggle_off_rows_after = toggle_df["toggle_off"].sum()

        logger.info(
            f"changed {toggle_on_rows_before - toggle_on_rows_after} "
            f"[{100 * (toggle_on_rows_before - toggle_on_rows_after) / toggle_on_rows_before:.1f}%] "
            f"rows from toggle_on True to False because toggle_change_settling_filter_seconds "
            f"= {cfg.toggle.toggle_change_settling_filter_seconds}",
        )
        logger.info(
            f"changed {toggle_off_rows_before - toggle_off_rows_after} "
            f"[{100 * (toggle_off_rows_before - toggle_off_rows_after) / toggle_off_rows_before:.1f}%] "
            f"rows from toggle_off True to False because toggle_change_settling_filter_seconds "
            f"= {cfg.toggle.toggle_change_settling_filter_seconds}",
        )

    if isinstance(toggle_df.index, pd.DatetimeIndex):
        return input_df.merge(
            toggle_df[["toggle_off", "toggle_on"]],
            how="left",
            left_index=True,
            right_index=True,
        )
    msg = f"toggle_df.index must be (or can be coerced to) a pd.DatetimeIndex, got {type(toggle_df.index)}"
    raise TypeError(msg)


@dataclass
class AssessmentInputs:
    wf_df: pd.DataFrame
    pc_per_ttype: dict[str, pd.DataFrame]
    cfg: WindUpConfig
    plot_cfg: PlotConfig
    pre_post_splitter: PrePostSplitter

    @classmethod
    def from_cfg(
        cls,
        cfg: WindUpConfig,
        plot_cfg: PlotConfig,
        *,
        scada_df: pd.DataFrame,
        metadata_df: pd.DataFrame,
        toggle_df: pd.DataFrame | None = None,
        reanalysis_datasets: list[ReanalysisDataset],
        cache_dir: Path | None = None,
    ) -> "AssessmentInputs":
        func = preprocess if cache_dir is None else with_pickle_cache(cache_dir / "preprocess.pickle")(preprocess)
        wf_df, pc_per_ttype = func(
            cfg=cfg,
            plot_cfg=plot_cfg,
            scada_df=scada_df,
            metadata_df=metadata_df,
            reanalysis_datasets=reanalysis_datasets,
        )
        cfg = add_smart_lat_long_to_cfg(md=metadata_df, cfg=cfg)
        pre_post_splitter = PrePostSplitter(cfg=cfg, toggle_df=toggle_df)
        return cls(
            wf_df=wf_df,
            pc_per_ttype=pc_per_ttype,
            cfg=cfg,
            plot_cfg=plot_cfg,
            pre_post_splitter=pre_post_splitter,
        )


def get_filtered_wf_df_and_cfg_with_latlongs(
    cfg: WindUpConfig,
    plot_cfg: PlotConfig,
    *,
    scada_df: pd.DataFrame,
    metadata_df: pd.DataFrame,
    reanalysis_datasets: list[ReanalysisDataset],
) -> tuple[pd.DataFrame, dict[str, pd.DataFrame], WindUpConfig]:
    wf_df, cfg = get_raw_scada_and_cfg_from_file(
        cfg=cfg,
        scada_df=scada_df,
        metadata_df=metadata_df,
        first_datetime_utc_start=cfg.lt_first_dt_utc_start,
        last_datetime_utc_start=cfg.analysis_last_dt_utc_start,
        plot_cfg=plot_cfg,
    )

    wf_df = filter_scada_df(wf_df, cfg=cfg, plot_cfg=plot_cfg)
    wf_df = add_reanalysis_data(wf_df, cfg=cfg, plot_cfg=plot_cfg, reanalysis_datasets=reanalysis_datasets)
    if cfg.optimize_northing_corrections:
        wf_df = auto_northing_corrections(wf_df, cfg=cfg, plot_cfg=plot_cfg)
    else:
        wf_df = apply_northing_corrections(wf_df, cfg=cfg, north_ref_wd_col=REANALYSIS_WD_COL, plot_cfg=plot_cfg)
    wf_df = add_wf_yawdir(wf_df, cfg=cfg)

    pc_per_ttype, rated_ws_per_ttype = calc_pc_and_rated_ws(
        cfg=cfg,
        wf_df=wf_df,
        x_col="WindSpeedMean",
        y_col="pw_clipped",
        x_bin_width=cfg.ws_bin_width / 2,
        plot_cfg=plot_cfg,
    )

    wf_df = add_ws_est(cfg=cfg, wf_df=wf_df, pc_per_ttype=pc_per_ttype, plot_cfg=plot_cfg)
    wf_df = add_waking_state(cfg=cfg, wf_df=wf_df, plot_cfg=plot_cfg)

    return wf_df, pc_per_ttype, cfg


def preprocess(
    cfg: WindUpConfig,
    plot_cfg: PlotConfig,
    *,
    scada_df: pd.DataFrame,
    metadata_df: pd.DataFrame,
    reanalysis_datasets: list[ReanalysisDataset],
) -> tuple[pd.DataFrame, dict[str, pd.DataFrame]]:
    logger.info(f"running wind_up analysis for {cfg.assessment_name}")

    wf_df, pc_per_ttype, cfg = get_filtered_wf_df_and_cfg_with_latlongs(
        cfg=cfg,
        plot_cfg=plot_cfg,
        scada_df=scada_df,
        metadata_df=metadata_df,
        reanalysis_datasets=reanalysis_datasets,
    )
    if cfg.ignore_turbine_anemometer_data:
        wf_df = wf_df.drop(columns=["ws_est_blend"])

    return wf_df, pc_per_ttype



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from tests.conftest import TEST_DATA_FLD
from wind_up.constants import RAW_DOWNTIME_S_COL, RAW_POWER_COL, RAW_YAWDIR_COL, TIMESTAMP_COL
from wind_up.models import WindUpConfig
from wind_up.optimize_northing import auto_northing_corrections, clip_wtg_north_table
from wind_up.reanalysis_data import ReanalysisDataset, add_reanalysis_data


def test_clip_wtg_north_table_entries_before() -> None:
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.Index(tstamps)
    wtg_df = pd.DataFrame(
        data={
            "ActivePowerMean": [3.14] * 3,
            "some_col_with_nans": [np.nan] * 3,
        },
        index=idx,
    )
    tstamps_for_wtg_north_table = [
        tstamps[0] - pd.Timedelta(days=2),
        tstamps[0] - pd.Timedelta(days=1),
        tstamps[-1],
        tstamps[-1] + pd.Timedelta(days=1),
    ]
    initial_wtg_north_table = pd.DataFrame(
        data={
            TIMESTAMP_COL: tstamps_for_wtg_north_table,
            "north_offset": list(range(len(tstamps_for_wtg_north_table))),
        },
    )
    expected_wtg_north_table = pd.DataFrame(
        data={
            TIMESTAMP_COL: [tstamps[0], tstamps[-1], tstamps[-1] + pd.Timedelta(days=1)],
            "north_offset": [1, 2, 3],
        },
    )
    actual_wtg_north_table = clip_wtg_north_table(initial_wtg_north_table, wtg_df=wtg_df)
    assert_frame_equal(actual_wtg_north_table, expected_wtg_north_table)


def test_clip_wtg_north_table_entry_exactly_at_start() -> None:
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.Index(tstamps)
    wtg_df = pd.DataFrame(
        data={
            "ActivePowerMean": [3.14] * 3,
            "some_col_with_nans": [np.nan] * 3,
        },
        index=idx,
    )
    tstamps_for_wtg_north_table = [
        tstamps[0] - pd.Timedelta(days=1),
        tstamps[0],
        tstamps[-1],
        tstamps[-1] + pd.Timedelta(days=1),
    ]
    initial_wtg_north_table = pd.DataFrame(
        data={
            TIMESTAMP_COL: tstamps_for_wtg_north_table,
            "north_offset": list(range(len(tstamps_for_wtg_north_table))),
        },
    )
    expected_wtg_north_table = pd.DataFrame(
        data={
            TIMESTAMP_COL: [tstamps[0], tstamps[-1], tstamps[-1] + pd.Timedelta(days=1)],
            "north_offset": [1, 2, 3],
        },
    )
    actual_wtg_north_table = clip_wtg_north_table(initial_wtg_north_table, wtg_df=wtg_df)
    assert_frame_equal(actual_wtg_north_table, expected_wtg_north_table)


def test_clip_wtg_north_table_entry_after_start() -> None:
    tstamps = pd.date_range(start="2021-01-01", tz="UTC", periods=3, freq="10min")
    idx = pd.Index(tstamps)
    wtg_df = pd.DataFrame(
        data={
            "ActivePowerMean": [3.14] * 3,
            "some_col_with_nans": [np.nan] * 3,
        },
        index=idx,
    )
    tstamps_for_wtg_north_table = [
        tstamps[-1] + pd.Timedelta(days=1),
    ]
    initial_wtg_north_table = pd.DataFrame(
        data={
            TIMESTAMP_COL: tstamps_for_wtg_north_table,
            "north_offset": list(range(len(tstamps_for_wtg_north_table))),
        },
    )
    expected_wtg_north_table = pd.DataFrame(
        data={
            TIMESTAMP_COL: [tstamps[0]],
            "north_offset": [0],
        },
    )
    actual_wtg_north_table = clip_wtg_north_table(initial_wtg_north_table, wtg_df=wtg_df)
    assert_frame_equal(actual_wtg_north_table, expected_wtg_north_table)


def test_auto_northing_corrections(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    cfg.lt_first_dt_utc_start = pd.Timestamp("2023-07-01 00:00:00", tz="UTC")
    cfg.analysis_last_dt_utc_start = pd.Timestamp("2023-07-31 23:50:00", tz="UTC")
    wf_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/Homer Wind Farm_July2023_scada_improved.parquet")
    reanalysis_datasets = [
        ReanalysisDataset(id=fp.stem, data=pd.read_parquet(fp))
        for fp in (TEST_DATA_FLD / "reanalysis" / "Homer Wind Farm").glob("*.parquet")
    ]
    wf_df = add_reanalysis_data(wf_df, cfg=cfg, plot_cfg=None, reanalysis_datasets=reanalysis_datasets)

    # add required columns not included in original test file
    wf_df[RAW_POWER_COL] = wf_df["ActivePowerMean"]
    wf_df[RAW_YAWDIR_COL] = wf_df["YawAngleMean"]
    wf_df[RAW_DOWNTIME_S_COL] = wf_df["ShutdownDuration"]

    northed_wf_df = auto_northing_corrections(wf_df, cfg=cfg, plot_cfg=None)

    median_yaw_before_northing = wf_df.groupby("TurbineName", observed=True)["YawAngleMean"].median()
    median_yaw_after_northing = northed_wf_df.groupby("TurbineName", observed=True)["YawAngleMean"].median()

    assert median_yaw_before_northing["HMR_T01"] == pytest.approx(191.0)
    assert median_yaw_before_northing["HMR_T02"] == pytest.approx(173.0)
    assert median_yaw_after_northing["HMR_T01"] == pytest.approx(269.6)
    assert median_yaw_after_northing["HMR_T02"] == pytest.approx(267.6)

    # try to mess up the yaw angles further and run again
    wf_df[RAW_YAWDIR_COL] = (wf_df[RAW_YAWDIR_COL] + 180) % 360

    # add a change point in for each turbine
    sign = 1
    for wtg_name in wf_df.index.unique(level="TurbineName"):
        idx = pd.IndexSlice[wtg_name, pd.Timestamp("2023-07-10 00:00:00+00:00") :]
        wf_df.loc[idx, RAW_YAWDIR_COL] = (wf_df.loc[idx, RAW_YAWDIR_COL] + sign * 30) % 360
        sign *= -1
    # add another change point for each turbine
    for wtg_name in wf_df.index.unique(level="TurbineName"):
        idx = pd.IndexSlice[wtg_name, pd.Timestamp("2023-07-20 00:00:00+00:00") :]
        wf_df.loc[idx, RAW_YAWDIR_COL] = (wf_df.loc[idx, RAW_YAWDIR_COL] + sign * 30) % 360

    wf_df["YawAngleMean"] = wf_df[RAW_YAWDIR_COL]

    northed_wf_df = auto_northing_corrections(wf_df, cfg=cfg, plot_cfg=None)

    median_yaw_before_northing = wf_df.groupby("TurbineName", observed=True)["YawAngleMean"].median()
    median_yaw_after_northing = northed_wf_df.groupby("TurbineName", observed=True)["YawAngleMean"].median()

    assert median_yaw_before_northing["HMR_T01"] == pytest.approx(178.0)
    assert median_yaw_before_northing["HMR_T02"] == pytest.approx(206.0)
    assert median_yaw_after_northing["HMR_T01"] == pytest.approx(269.2)
    assert median_yaw_after_northing["HMR_T02"] == pytest.approx(269.2)



from pathlib import Path

import pandas as pd
from pandas.testing import assert_frame_equal

from wind_up.combine_results import combine_results


def calc_expected_combine_results(trdf: pd.DataFrame) -> pd.DataFrame:
    p50_uplift = (trdf["uplift_frc"] * (1 / trdf["unc_one_sigma_frc"] ** 2)).sum() / (
        1 / trdf["unc_one_sigma_frc"] ** 2
    ).sum()
    sigma_correlated = (1 / trdf["unc_one_sigma_frc"]).sum() / (1 / trdf["unc_one_sigma_frc"] ** 2).sum()
    sigma_independent = (
        (
            (
                trdf["unc_one_sigma_frc"]
                * 1
                / (trdf["unc_one_sigma_frc"] ** 2)
                / (1 / (trdf["unc_one_sigma_frc"] ** 2)).sum()
            )
            ** 2
        ).sum()
    ) ** 0.5
    sigma_test = (sigma_independent + sigma_correlated) / 2
    return pd.DataFrame(
        data={
            "test_wtg": ["test1"],
            "p50_uplift": [p50_uplift],
            "sigma": [sigma_test],
            "sigma_test": [sigma_test],
            "sigma_uncorr": [sigma_independent],
            "sigma_corr": [sigma_correlated],
        },
    )


def test_combine_two_refs() -> None:
    trdf = pd.DataFrame(
        data={
            "test_wtg": ["test1", "test1"],
            "ref": ["ref1", "ref2"],
            "uplift_frc": [0.02, 0.02],
            "unc_one_sigma_frc": [0.02, 0.02],
        },
    )
    edf = calc_expected_combine_results(trdf)
    tdf = combine_results(trdf=trdf, auto_choose_refs=False)
    tdf = tdf[edf.columns.tolist()]
    assert_frame_equal(edf, tdf)


def test_combine_three_refs() -> None:
    trdf = pd.DataFrame(
        data={
            "test_wtg": ["test1", "test1", "test1"],
            "ref": ["ref1", "ref2", "ref3"],
            "uplift_frc": [0.01, 0.02, 0.03],
            "unc_one_sigma_frc": [0.03, 0.02, 0.01],
        },
    )
    edf = calc_expected_combine_results(trdf)

    tdf = combine_results(trdf=trdf, auto_choose_refs=False)

    tdf = tdf[edf.columns.tolist()]
    assert_frame_equal(edf, tdf)


def test_brt_t16_pitch() -> None:
    trdf = pd.read_csv(Path(__file__).parents[0] / "test_data/trdf_BRT_T16_pitch_Sep23.csv", index_col=0)
    edf = pd.read_csv(Path(__file__).parents[0] / "test_data/tdf_BRT_T16_pitch.csv", index_col=0)
    tdf = combine_results(trdf=trdf)
    assert_frame_equal(edf, tdf)


def test_brt_t16_pitch_no_auto_choose() -> None:
    trdf = pd.read_csv(Path(__file__).parents[0] / "test_data/trdf_BRT_T16_pitch_Sep23.csv", index_col=0)
    edf = pd.read_csv(Path(__file__).parents[0] / "test_data/tdf_BRT_T16_pitch_no_auto_choose.csv", index_col=0)
    tdf = combine_results(trdf=trdf, auto_choose_refs=False)
    assert_frame_equal(edf, tdf)


def test_brt_t16_pitch_exclude_refs() -> None:
    trdf = pd.read_csv(Path(__file__).parents[0] / "test_data/trdf_BRT_T16_pitch_Sep23.csv", index_col=0)
    edf = pd.read_csv(Path(__file__).parents[0] / "test_data/tdf_BRT_T16_pitch_exclude_refs.csv", index_col=0)
    # all but one ref excluded
    tdf = combine_results(trdf=trdf, auto_choose_refs=False, exclude_refs=["BRT_T02", "BRT_T03", "BRT_T04", "BRT_T14"])
    assert_frame_equal(edf, tdf)



from pathlib import Path

import toml

import wind_up
from wind_up.constants import PROJECTROOT_DIR


def test_versions_match() -> None:
    with Path.open(PROJECTROOT_DIR / "pyproject.toml") as toml_file:
        pyproject_data = toml.load(toml_file)
    package_version_from_toml = pyproject_data["project"]["version"]
    package_version_from_init = wind_up.__version__
    assert package_version_from_toml == package_version_from_init



import pickle
from collections.abc import Callable
from pathlib import Path
from typing import Any

import pandas as pd

from wind_up.result_manager import result_manager


def with_parquet_cache(fp: Path, *, use_cache: bool = True) -> Callable:
    def wrap(func: Callable[..., pd.DataFrame]) -> Callable[..., pd.DataFrame]:
        def wrapped_f(*a: Any, **kw: Any) -> pd.DataFrame:  # noqa
            if not Path(fp).is_file() or not use_cache:
                func(*a, **kw).to_parquet(fp)
            return pd.read_parquet(fp)

        return wrapped_f

    return wrap


def with_pickle_cache(fp: Path, *, use_cache: bool = True) -> Callable:
    def wrap(func: Callable[..., Any]) -> Callable[..., Any]:
        def wrapped_f(*a: Any, **kw: Any) -> Any:  # noqa
            if not Path(fp).is_file() or not use_cache or Path(fp).stat().st_size == 0:
                with Path.open(fp, "wb") as f:
                    pickle.dump(func(*a, **kw), f)
            with Path.open(fp, "rb") as f:
                result_manager.warning(f"loading cached pickle {fp}")
                return pickle.load(f)

        return wrapped_f

    return wrap
